//---------------------------------------------------------------------------
/*
ADPCM Program for Windows

 Copyright 2020 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page: http://www.silogic.com/trains/ADPCM.html
//---------------------------------------------------------------------------
/*
from the sox man page:

  https://github.com/ldfaiztt/CSE332/blob/master/Project%201/sox12181/sox.txt

		 ADPCM is a form of sound compression that has a good  compro-
		 mise  between	good  sound quality and fast encoding/decoding
		 time.	It is used for telephone sound compression and	places
		 were full fidelity is not as important.  When uncompressed it
		 has roughly the precision of 16-bit PCM audio.	 Popular  ver-
		 sion of ADPCM include G.726, MS ADPCM, and IMA ADPCM.	The -a
		 flag has different meanings in different file	handlers.   In
		 .wav  files  it  represents  MS ADPCM files, in all others it
		 means G.726 ADPCM.  IMA ADPCM is a  specific  form  of	 ADPCM
		 compression,  slightly	 simpler  and  slightly lower fidelity
		 than Microsofts flavor of ADPCM.  IMA ADPCM is  also	called
		 DVI ADPCM.

*/

//---------------------------------------------------------------------------
#ifdef DRAGDROP
/*
Drag and Drop uses TDragObject which works in both BCB 4.0 and CB 10.1.
If this program were only for CB 10.1, we could have used TDragObjectEx which
is deleted automatically when the drop is complete. With TDragObject, I have to
create it in the constructor and then delete it in the destructor.

Drag and Drop does not correctly scroll the sound list. I removed this
feature until I figure out how to scroll correctly.
*/
#endif
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <tchar.h>
#include <windows.h>
#include <mmsystem.h>
#include <stdio.h>
#include <assert.h>

#include <dir.h>
#include <shellapi.h>
#include <Filectrl.hpp>

//#define DRAGDROP

#include "ADPCM_Player.h"
#include "ADPCM_Info.h"
#include "Edit_the_Clip.h"
#include "Edit_the_Index.h"
#include "Edit_the_Softkeys.h"
#include "Edit_the_Scale_Factor.h"
#include "Edit_the_Engine_Parameters.h"

/*
Code from :
https://www.planet-source-code.com/vb/scripts/ShowCode.asp?txtCodeId=4422&lngWId=3
Playing Audio in Windows using waveOut Interface
	Email
Submitted on: 	1/4/2015 1:42:00 AM
By: 	David Overton (from psc cd)
/*
 * some good values for block size and count
 */
#define BLOCK_SIZE 8192
#define BLOCK_COUNT 20
/*
 * function prototypes
 */ 
static void CALLBACK waveOutProc(HWAVEOUT, UINT, DWORD, DWORD, DWORD);
static WAVEHDR* allocateBlocks(int size, int count);
static void freeBlocks(WAVEHDR* blockArray);
static void writeAudio(HWAVEOUT hWaveOut, LPSTR data, unsigned long size);

/*
* module level variables
*/
WAVEHDR* waveBlocks;
static CRITICAL_SECTION waveCriticalSection;
volatile int waveFreeBlockCount;
int waveCurrentBlock;
HWAVEOUT hWaveOut; /* device handle */

void __fastcall PlaySegment(HWAVEOUT hWaveOut, HANDLE hSoundfile, int StartAddr, int EndAddr, int DisplayAddr, ImaState & state);
void writeSample(HWAVEOUT hWaveOut, short Value);
int BufIndex;
static short sBuffer2048[2048];		// 2048 samples of 16 bit PCM decompressed buffer for 16 bit samples

// structure containing the sample rates in various forms
SRSetting SR[6] = {{HZ5512, TEXT("5512"),TEXT("5.512"),5512},{HZ8000, TEXT("8000"),TEXT("8.000"),8000},
		{HZ11025, TEXT("11025"),TEXT("11.025"),11025},{HZ16000, TEXT("16000"),TEXT("16.000"),16000},
		{HZ22050, TEXT("22050"),TEXT("22.050"),22050},{HZ44100, TEXT("44100"),TEXT("44.100"),44100}};

bool Stopem;
/*
//---------------------------------------------------------------------------
// Two tables to map the "data" byte in the index to the Sample Rate.
//
// PS2 engines use the range 0x00-0x7F - the HO bit is always
// zero.  Data format for PS2 is always 4 bit IMA/DVI ADPCM.
//
// PS3 use the range 0x00-0xFF. Data format is 4 bit IMA/DVI ADPCM if the HO bit
// is '0', and is 24 bit big endian signed PCM if the HO bit is a '1'.
//
signed char SampleRatePS2[128] = {
// -1 = unknown, 0 = 5,512 Hz, 2 = 11,025 Hz, 4 = 22,050 Hz
//
//     Low Nybble/HO 2 bits - LO 2 bits always "11"
//   0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
//   X  X  X  0  X  X  X  1  X  X  X  2  X  X  X  3     //High Nybble - ADPCM
	 0,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 2,-1,-1,-1, 2,	// 0    Engine
	-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,	// 1    Accent
	-1,-1,-1, 4,-1,-1,-1, 4,-1,-1,-1, 2,-1,-1,-1, 2,    // 2    Horn/Whistle
	-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,	// 3    Bell
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0,-1,-1,-1, 0,	// 4    Engine
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0,-1,-1,-1, 0,	// 5    Accent
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,	// 6    not used
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1		// 7    not used
	};                                                  //High Nybble
signed char SampleRatePS3[256] = {
// -1 = unknown, 0 = 5,512 Hz, 2 = 11,025 Hz, 4 = 22,050 Hz
//
//     Low Nybble/HO 2 bits - LO 2 bits always "11"
//   0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
//   X  X  X  0  X  X  X  1  X  X  X  2  X  X  X  3     //High Nybble - ADPCM
	 2,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,	// 0    Engine
	-1,-1,-1, 4,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,	// 1    Accent
	-1,-1,-1, 4,-1,-1,-1, 4,-1,-1,-1, 2,-1,-1,-1, 2,    // 2    Horn/Whistle
	-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 2,	// 3    Bell
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0,-1,-1,-1, 0,	// 4    Engine
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0,-1,-1,-1, 0,	// 5    Accent
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,	// 6    not used
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,	// 7    not used
//     Low Nybble/HO 2 bits - LO 2 bits always "11"
//   0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
//   X  X  X  0  X  X  X  1  X  X  X  2  X  X  X  3     //High Nybble - 24 bit PCM
	 2,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 0,-1,-1,-1, 0,	// 8    Engine
	-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 0,-1,-1,-1, 0,	// 9    Accent
	-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 0,-1,-1,-1, 0,	// A    Horn/Whistle
	-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1, 0,-1,-1,-1, 0,	// B    Bell
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0,-1,-1,-1, 0,	// C    Engine
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0,-1,-1,-1, 0,	// D    Accent
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,	// E    not used
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1		// F    not used
	};
*/
/*
PS2: examples

0x00	0000 0000   4 bit ADPCM 5,512 Hz confirmed with voice 0x0B50 SD70ACE
0x03	0000 0011   4 bit ADPCM 5,512 Hz

0x4B	0100 1011   4 bit ADPCM 5,512 Hz confirmed with steam release
0x5B	0101 1011   4 bit ADPCM 5,512 Hz confirmed with voice, confirmed with horn (SD70ACE)
0x07	0000 0111   4 bit ADPCM 5,512 Hz
0x4F    0100 1111   4 bit ADPCM 5,512 Hz
0x5F    0101 1111   4 bit ADPCM 5,512 Hz

0x0B	0000 1011   4 bit ADPCM 11,025 Hz
0x13	0001 0011   4 bit ADPCM 11,025 Hz (PS2)
0x1B	0001 1011   4 bit ADPCM 11,025 Hz
0x2B	0010 1011   4 bit ADPCM 11,025 Hz confirmed with horn 0x0380 SD70ACE

0x0F	0000 1111   4 bit ADPCM 11,025 Hz confirmed with engine rumble
0x1F	0001 1111   4 bit ADPCM 11,025 Hz confirmed with brake sequeal, confirmed with clickity clack
0x2F	0010 1111   4 bit ADPCM 11,025 Hz confirmed with 0-8-0 whistle
0x3F    0011 1111   4 bit ADPCM 11,025 Hz confirmed with 0-8-0 bell
0x23	0010 0011   4 bit ADPCM 11,025 Hz Whistle

0x27	0010 0111   4 bit ADPCM 22,050 Hz Whistle
0x37	0011 0111   4 bit ADPCM 22,050 Hz Bell

PS3: examples

0x5B	0101 1011   4 bit ADPCM 5,512 Hz confirmed with voice "how's she been running"

0x00	0000 0000	4 bit ADPCM 11,025 Hz (PS3)  (0x00 is 5,512 Hz in PS2)
0x03	0000 0011   4 bit ADPCM 11,025 Hz (PS3)  (0x03 is 5,512 Hz in PS2)
0x07	0000 0111   4 bit ADPCM 11,025 Hz (PS3)
0x1B	0001 1011	4 bit ADPCM 11,025 Hz (PS3)

0x13	0001 0011   4 bit ADPCM 22,050 Hz (PS3) confirmed with GP7 0x0820 "all aboard" (0x13 is 11,025 Hz in PS2)

0xDB	1101 1011	24 bit signed PCM 2,756 Hz (not supported by this program)

0x9B	1001 1011   24 bit signed PCM 5,512 Hz confirmed with GP7 0x710 voice
0x9F    1001 1111   24 bit signed PCM 5,512 Hz confirmed with flange squeal
0xAB	1010 1011   24 bit signed PCM 5,512 Hz (0x03B0 or 0x10E0 (last index) on GP7) confirmed with GP7 horn 0x01010
0xBF    1011 1111   24 bit signed PCM 5,512 Hz confirmed with GP7 bell  0xAC0

0x83	1000 0011   24 bit signed PCM 11,025 Hz
0x87	1000 0111   24 bit signed PCM 11,025 Hz
0x93	1001 0011   24 bit signed PCM 11,025 Hz
0x97	1001 0111   24 bit signed PCM 11,025 Hz
0xA7	1010 0111   24 bit signed PCM 11,025 Hz
0xA3	1010 0011   24 bit signed PCM 11,025 Hz

PS3 Not in table:

0xDB	1101 1011   24 bit signed PCM < 5,512 Hz (GP7 0x4B0)
*/

/*
Data Value:
bit 7 (high order bit)  0 = IMA/DVI 4 bit ADPCM clip
						1 = 24 bit signed big endian PCM clip (PS3 sound files only)
bit 6 - possibly, with bits 7 & 3, sets the sample rate and format per the table below.
bit 5-4 - appears to connect the clip with the volume control that controls it:
	00 - Engine sound clip
	01 - Accent sound clip
	10 - Horn/Whistle sound clip
	11 - Bell sound clip
bit 3 - possibly, with bits 7 & 6, sets the sample rate and format per the table below.
bit 2 - unknown
bit 1-0 - unknown, either "00" or "11", but mostly "11".

--------------------------------
Bit	7  	6  	3

	0 	0 	0 	22050 Hz ADPCM

	0 	0 	1 	11025 Hz ADPCM

	0 	1 	0 	not seen

	0 	1 	1 	5512 Hz ADPCM

	1 	0 	0 	11025 Hz PCM

	1 	0 	1 	5512 Hz PCM

	1 	1 	0 	not seen

	1 	1 	1 	not seen


*/
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TADPCM_Main *ADPCM_Main;
//---------------------------------------------------------------------------
__fastcall TADPCM_Main::TADPCM_Main(TComponent* Owner)
	: TForm(Owner)
{
SS = SPLASH_ON;		// Splash Window if == 1

// User Registry Data
RegFileRootKey = "Software\\Remote Control\\ADPCM";

StringGrid1->Cells[ICLIPNO][0] = "Clip #";
StringGrid1->Cells[ISTARTADDR][0] = "Start";
StringGrid1->Cells[IENDREPADDR][0] = "End Rep";	// end of repeatable portion of clip
StringGrid1->Cells[ISTARTREPADDR][0] = "Start Rep";	// start of repeatable portion of clip
StringGrid1->Cells[ISMINIT][0] = "SM Init";	// initial values for IMA ADPCM state machine when playing the Repeat Segment
											// 0xYYZZZZ : YY=index, ZZZZ=predValue
StringGrid1->Cells[IENDADDR][0] = "End";
StringGrid1->Cells[ILENGTH][0] = "Length";
StringGrid1->Cells[IDATABYTE][0] = "Data";

StringGrid1->RowCount =  (0x10f0 - 0x0100) / 16;
// plus one for the Title Row
StringGrid1->RowCount += 1;
Decoded = false;
IndexFormat = false;
//ExportOneClipFlag = false;
RecordFlag = false;
iDebug = 0;

hWaveOut = NULL;
//
// generate the version number and display it
//
DWORD  verHandle = NULL;
UINT   size      = 0;
LPBYTE lpBuffer  = NULL;
DWORD  verSize   = GetFileVersionInfoSize( Application->ExeName.c_str(), &verHandle);

if (verSize != NULL) {
	char Version[128];	// holds program version number
    LPSTR verData = new char[verSize];
    assert (verData != NULL);
    if (GetFileVersionInfo( Application->ExeName.c_str(), verHandle, verSize, verData)) {
        if (VerQueryValue(verData, TEXT("\\"),(VOID FAR* FAR*)&lpBuffer,&size)) {
			if (size) {
                VS_FIXEDFILEINFO *verInfo = (VS_FIXEDFILEINFO *)lpBuffer;
                if (verInfo->dwSignature == 0xfeef04bd) {
                    // Doesn't matter if you are on 32 bit or 64 bit,
					// DWORD is always 32 bits, so first two revision numbers
					// come from dwFileVersionMS, last two come from dwFileVersionLS
					snprintf(Version, sizeof Version, "V%d.%d.%d.%d (%s)",
					( verInfo->dwFileVersionMS >> 16 ) & 0xffff,
                    ( verInfo->dwFileVersionMS >>  0 ) & 0xffff,
                    ( verInfo->dwFileVersionLS >> 16 ) & 0xffff,
                    ( verInfo->dwFileVersionLS >>  0 ) & 0xffff,
                    __DATE__);
                    VersionLabel->Caption = Version;
                	}
            	}
        	}
    	}
    delete[] verData;
	} else {
    // Retrieve the system error message for the last-error code
    LPVOID lpMsgBuf;
    DWORD dw = GetLastError();

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );
    // Display the error message
    MessageBox(NULL, (LPCTSTR)lpMsgBuf, TEXT("Error"), MB_OK);
    LocalFree(lpMsgBuf);
    }

//
randomize();
EEPROMSize = 0;
EngineEEPROMSize->Caption = "";
AvailableSpace = 0;

// Experimental - use Drop Down Menu style in CB10.1
#if __BORLANDC__ >= 0x720
enum DECLSPEC_DENUM TButtonStyle : unsigned char { bsPushButton, bsCommandLink, bsSplitButton };
// bsSplitButton defined in Vcl.StdCtrls.hpp is really not defined
BrowseforFileButton->Style = bsSplitButton;
BrowseforFileButton->DropDownMenu = FileNamePopupMenu;

PlayClipButton->Style = bsSplitButton;
PlayClipButton->DropDownMenu = ADPCMPopupMenu;
#endif

//
tm = new TMyDragObject();
Edit_Engine_Parameters = NULL;
Edit_Scale_Factor = NULL;
}
//---------------------------------------------------------------------------
__fastcall TADPCM_Main::~TADPCM_Main()
{
delete tm;
}
//---------------------------------------------------------------------------
int MyGetFileSizeBytes(HANDLE hFile)
{
// return the file size in bytes
return GetFileSize(hFile, NULL);
}
//---------------------------------------------------------------------------

int getEEPROMSize(HANDLE hFile)
{
// based on the size of the Sound File, return the probable EEPROM memory size in bytes
const int FileSize = MyGetFileSizeBytes(hFile);	// number of bytes in the file
if (FileSize <= 0 ) return FileSize;	// error
if (ADPCM_Main->PSGroup->ItemIndex == PS3) {    // if it has 24 bit PCM clips, then it's PS3
	// force 4MB EEPROM
	ADPCM_Main->EngineEEPROMSize->Caption = "4 MB";	// must be 4MB if PS3
	return 0x3A0000;	// 4 MB - when uploaded though, we get 3,801,088 bytes (0x3A0000)
	}
if (FileSize <= 0x100000) {
	ADPCM_Main->EngineEEPROMSize->Caption = "1 MB";
	return 0x100000;	// 1 MB - when uploaded, we get 1,048,576 bytes (0x100000)
	}
if (FileSize <= 0x200000) {
	ADPCM_Main->EngineEEPROMSize->Caption = "2 MB";
	return 0x200000;	// 2 MB - when uploaded, we get 2,097,152 bytes (0x200000)
	}
// otherwise, its 4MB
ADPCM_Main->EngineEEPROMSize->Caption = "4 MB";
ADPCM_Main->PSGroup->ItemIndex = PS3;	// if memory is 4MB, it's PS3
return 0x3A0000;	// 4 MB - when uploaded though, we get 3,801,088 bytes (0x3A0000)
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::getFileDetails(HANDLE hFile)
{
// some file details
getFileString(hFile, 0x1DD8, 32, "Engine Number : ");
getFileString(hFile, 0x1DF8, 32, "Railroad : ");
getFileString(hFile, 0x1E18, 32, "Engine Name : ");
getFileString(hFile, 0x1E38, 32, "Unknown Field : ");
getFileString(hFile, 0x1E58, 32, "Hex Code File Name : ");
getFileString(hFile, 0x1E78, 32, "Unknown Field : ");
getFileString(hFile, 0x1E98, 32, "Internal Sound File Name : ");
getFileString(hFile, 0x1EB8, 32, "Inspector : ");
getFileString(hFile, 0x603D, 32, "New Engine Name : ");
//
SoundFileSize = MyGetFileSizeBytes(hFile);	// number of bytes in the file
MyShowMessage("Sound File Size in Bytes = " + String(SoundFileSize) + " (0x" + IntToHex(SoundFileSize, IDIGITSTODISPLAY) + ")", false);

EEPROMSize = getEEPROMSize(hFile);	// in bytes
// in the case where the Sound File Size is a multiple of 1 MB, look
// for 0xFF fill at the end of the file
if ((SoundFileSize % 0x100000) == 0 ) {
	// starting at EOF, read backwards looking for first byte that
	// is not 0xFF
	unsigned char c;
	DWORD readBytes;
	int i;
	for (i = 1 ; i <= SoundFileSize ; i++){		// count in bytes
		// set to EOF-1 and so on
		unsigned long lretv = SetFilePointer(hFile, - i, NULL, FILE_END);
		int retv = ReadFile(hFile, &c, 1, &readBytes, NULL);
		if (retv != true) {
			MyShowMessage("getFileDetails: didn't read byte at " + String(lretv), false);
			}
		if (c != 0xFF) break;
		}
	SoundFileSize = SoundFileSize - i;
	SoundFileSize++;
	}
AvailableSpace = EEPROMSize - SoundFileSize;	// in bytes

MyShowMessage("EEPROM Size in Bytes = " + String(EEPROMSize) + " (0x" + IntToHex(EEPROMSize, IDIGITSTODISPLAY) + ")", false); //TODO
MyShowMessage("Used Space in Bytes = " + String(SoundFileSize) + " (0x" + IntToHex(SoundFileSize, IDIGITSTODISPLAY) + ")", false);
MyShowMessage("Available Space in Bytes = " + String(AvailableSpace), false);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::getFileString(HANDLE hFile, long offsetBytes, int LenBytes, String Caption)
{
assert(LenBytes < 256);
//
// display various strings if they exist in the sound file
//
const unsigned long lretv = SetFilePointer(hFile, offsetBytes, NULL, FILE_BEGIN);
unsigned char InternalString[257];
DWORD readBytes;
int retv = ReadFile(hFile, InternalString, LenBytes, &readBytes, NULL);
if (retv != true) {
	MyShowMessage("getFileString: didn't read string at " + String(lretv), false);
	}
if (InternalString[0] != 0xFF) {
	String IS = "";
	for (DWORD i = 0 ; i < readBytes ; i++) {	// remove imbedded zeros
		if (InternalString[i] != 0x00) IS += String((char)InternalString[i]);
		}
	MyShowMessage(Caption +  IS, false);
	}

}
//---------------------------------------------------------------------------

static void CALLBACK waveOutProc(HWAVEOUT hWaveOut, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
/*
 * pointer to free block counter
 */
int* freeBlockCounter = (int*)dwInstance;
/*
 * ignore calls that occur due to opening and closing the device.
 */
if(uMsg != WOM_DONE) return;
EnterCriticalSection(&waveCriticalSection);
(*freeBlockCounter)++;
LeaveCriticalSection(&waveCriticalSection);
}
//---------------------------------------------------------------------------

WAVEHDR* allocateBlocks(int size, int count)
{
unsigned char * buffer;

WAVEHDR* blocks;
const DWORD totalBufferSize = (size + sizeof(WAVEHDR)) * count;
/*
 * allocate memory for the entire set in one go
 */
if((buffer = (unsigned char *) HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, totalBufferSize)) == NULL) {
	ADPCM_Main->MyShowMessage("allocateBlocks: Memory allocation error", true);
	return (NULL);
	}
/*
 * and set up the pointers to each bit
 */
blocks = (WAVEHDR*)buffer;
buffer += sizeof(WAVEHDR) * count;
for(int i = 0; i < count; i++) {
	blocks[i].dwBufferLength = size;
	blocks[i].lpData = (LPSTR) buffer;
	buffer += size;
	}
return blocks;
}
//---------------------------------------------------------------------------

void freeBlocks(WAVEHDR* blockArray)
{
/*
 * and this is why allocateBlocks works the way it does
 */
HeapFree(GetProcessHeap(), 0, blockArray);
}
//---------------------------------------------------------------------------

void writeAudio(HWAVEOUT hWaveOut, LPSTR data, unsigned long size)
{
WAVEHDR* current;
int remain;
current = &waveBlocks[waveCurrentBlock];
ADPCM_Main->BuffersSent++;
// write to disk file
if (ADPCM_Main->RecordFlag) { 	// records as 16 bit little endian signed PCM
	unsigned long byteswritten;
	WriteFile(ADPCM_Main->hRawFile, data, size, &byteswritten, NULL);
	if (byteswritten != size) {
		ADPCM_Main->MyShowMessage("writeAudio: did not record correct number of bytes", true);
		}
	}
while(size > 0) {
	/*
	 * first make sure the header we're going to use is unprepared
	 */
	if(current->dwFlags & WHDR_PREPARED)
		if (waveOutUnprepareHeader(hWaveOut, current, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) {
				ADPCM_Main->MyShowMessage("writeAudio: waveOutUnprepareHeader Failure", true);
				}
	if(size < (unsigned long)(BLOCK_SIZE - current->dwUser)) {
		memcpy(current->lpData + current->dwUser, data, size);
		current->dwUser += size;
		break;
		}
	remain = BLOCK_SIZE - current->dwUser;
	memcpy(current->lpData + current->dwUser, data, remain);
	size -= remain;
	data += remain;
	current->dwBufferLength = BLOCK_SIZE;
	if (waveOutPrepareHeader(hWaveOut, current, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) {
		ADPCM_Main->MyShowMessage("writeAudio: waveOutPrepareHeader Failure", true);
		}
	if (waveOutWrite(hWaveOut, current, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) {
		ADPCM_Main->MyShowMessage("writeAudio: waveOutWrite Failure", true);
		}
	EnterCriticalSection(&waveCriticalSection);
	waveFreeBlockCount--;
	LeaveCriticalSection(&waveCriticalSection);
	/*
	 * wait for a block to become free
	 */
	while(!waveFreeBlockCount)
		Sleep(10);
	/*
	 * point to the next block
	 */
	waveCurrentBlock++;
	waveCurrentBlock %= BLOCK_COUNT;
	current = &waveBlocks[waveCurrentBlock];
	current->dwUser = 0;
	}
}
//---------------------------------------------------------------------------
//
// From: Algorithm Alley
// Inside IMA ADPCM    by Tim Kientzle
// http://www.drdobbs.com/database/algorithm-alley/184410326
// https://wiki.multimedia.cx/index.php?title=IMA_ADPCM

static const short _stepSizeTable[89] = {
   7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 19, 21, 23, 25, 28, 31, 34,
   37, 41, 45, 50, 55, 60, 66, 73, 80, 88, 97, 107, 118, 130, 143,
   157, 173, 190, 209, 230, 253, 279, 307, 337, 371, 408, 449, 494,
   544, 598, 658, 724, 796, 876, 963, 1060, 1166, 1282, 1411, 1552,
   1707, 1878, 2066, 2272, 2499, 2749, 3024, 3327, 3660, 4026,
   4428, 4871, 5358, 5894, 6484, 7132, 7845, 8630, 9493, 10442,
   11487, 12635, 13899, 15289, 16818, 18500, 20350, 22385, 24623,
   27086, 29794, 32767
};

static const short indexAdjustTable[16] = {
   -1, -1, -1, -1,  // +0 - +3, decrease the step size
    2, 4, 6, 8,     // +4 - +7, increase the step size
   -1, -1, -1, -1,  // -0 - -3, decrease the step size
    2, 4, 6, 8,     // -4 - -7, increase the step size
};

//---------------------------------------------------------------------------
#ifdef CLICKING
// was declared short
int ImaAdpcmDecode(unsigned char deltaCode, ImaState &state) {
#else
short ImaAdpcmDecode(unsigned char deltaCode, ImaState &state) {
#endif
   // delta Code is always 4 bits long
   // Get the current step size
   const short step = _stepSizeTable[state.step];
   // Construct the difference by scaling the current step size
   // This is approximately: difference = (deltaCode+.5)*step/4
   short difference = step>>3;
   // added () around the right shifts
   if ( deltaCode & 0x01 ) difference += (step>>2);
   if ( deltaCode & 0x02 ) difference += (step>>1);
   if ( deltaCode & 0x04 ) difference += step;
   if ( deltaCode & 0x08 ) difference = -difference;
#ifdef CLICKING
   // Build the new sample
   int ipredValue =  state.predValue;
   ipredValue += difference;			// NOTE: this value can get bigger than a 16 bit short can hold
										// so result MUST be put into an int
                                        // See below, this failed on the 30-2198-1 sound file clip 122
   if (ipredValue > 32767) {
		//ADPCM_Main->MyShowMessage("previous Value > 32767", false);
		ipredValue = 32767;	// keep within a 16 bit signed value
        ADPCM_Main->Overflow->Visible = true;
        ADPCM_Main->Overflow->Checked = true;
		} else {
		if (ipredValue < -32768) {
			ipredValue = -32768;
	        ADPCM_Main->Overflow->Visible = true;
    	    ADPCM_Main->Overflow->Checked = true;
			//ADPCM_Main->MyShowMessage("previous Value < -32768", false);
			}
		}
   state.predValue = ipredValue;
#else
   // Build the new sample
   state.predValue += difference;		// NOTE: this value may get bigger than a 16 bit short can hold
										// I saw this on one sound file but I don't remember which one
                                        // I changed the code (see above) to use an int and then
                                        // check for overflow but that failed on 30-2198-1 sound file
                                        // with a lot of clicking (clip 122)
#endif
   //
   // Update the step for the next sample
   state.step += indexAdjustTable[deltaCode];
   if (state.step < 0) state.step = 0;
   else if (state.step > 88) state.step = 88;
 
   return state.predValue; 	// this return value will always fit into a 16 bit short
}
//---------------------------------------------------------------------------
#ifdef NOTUSED
// This function is not used in the ADPCM program
unsigned char ImaAdpcmEncode(int sample, ImaState &state) {
   int diff = sample - state.predValue;
   int step = _stepSizeTable[state.step];
   unsigned char deltaCode = 0;
 
   // Set sign bit
   if (diff < 0) { deltaCode = 8; diff = -diff; }
 
   // This is essentially deltaCode = (diff<<2)/step,
   // except the roundoff is handled differently.
   if ( diff >= step ) {  deltaCode |= 4;  diff -= step;  }
   step >>= 1;
   if ( diff >= step ) {  deltaCode |= 2;  diff -= step;  }
   step >>= 1;
   if ( diff >= step ) {  deltaCode |= 1;  diff -= step;  }

   ImaAdpcmDecode(deltaCode, state);  // update state
   return (unsigned char) deltaCode;
}
#endif
//---------------------------------------------------------------------------

int BufferToAddress(unsigned char * buffer)
{
// convert a three character string to a byte address
// based on sporadic's posting and corrected for 32K frame
//
// Note that the address in the actual file is a nybble reference
//
#ifdef NYBBLEADDRESSES
return (((buffer[0] << 16) | (buffer[1] << 8) | buffer[2]));
#else
////return 0x8000L + ((buffer[0] - 1) * 0x8000L) + (((buffer[1] * 256) + buffer[2]) >> 1);
////return (buffer[0] * 0x8000L) + (((buffer[1] << 8) + buffer[2]) >> 1);
return (((buffer[0] << 16) | (buffer[1] << 8) | buffer[2]) >> 1);
#endif
}
//---------------------------------------------------------------------------

void AddressToBuffer(int Addr, unsigned char * buffer)
{
// convert a byte address into a three character string
// based on sporadic's posting and corrected for 32K frame
//
#ifdef NYBBLEADDRESSES
buffer[0] = (Addr & 0xFF0000) >> 16;
buffer[1] = (Addr & 0x00FF00) >> 8;
buffer[2] = (Addr & 0x0000FF);
#else
//  0x8000L + ((buffer[0] - 1) * 0x8000L) + (((buffer[1] * 256) + buffer[2]) >> 1);
buffer[0] = (Addr & 0x7F8000) >> 15;
buffer[1] = (Addr & 0x7F80 ) >> 7;
buffer[2] = (Addr & 0x7F) << 1;
// special case : correct 0xFFFFFE to 0xFFFFFF
if (buffer[0] == 0xFF && buffer[1] == 0xFF && buffer[2] == 0xFE) buffer[2] = 0xFF;
#endif
return;
}
//---------------------------------------------------------------------------
/*
This is not a good idea

bool AdjustLength(SFIndex &Index) {
//
// Length of PCM clip must always contain an integral number of samples (3 bytes)
// Length of ADPCM clip is always an integral number of bytes
//
// For 24bit PCM clips, round the length down to a multiple of 3
if (Index.databyte & PCM24MASK) {
	// TODO evaluate correction
	if (Index.ClipLength % 3 != 0) Index.ClipLength--;
	if (Index.ClipLength % 3 != 0) Index.ClipLength--;
	if (Index.ClipRepLength % 3 != 0) Index.ClipRepLength--;
	if (Index.ClipRepLength % 3 != 0) Index.ClipRepLength--;	
	}
return true;
}
*/
//---------------------------------------------------------------------------

bool ParseIndex(SFIndex &Index, HANDLE hParseFile, int clip)
{
// get a 16 byte index entry and parse it into the SFIndex structure named Index
// clip number is 1 based
SetFilePointer(hParseFile, 0x0100 + ((clip - 1) * 16), NULL, FILE_BEGIN);
unsigned char buffer16[16];
DWORD readBytes;
// read 16 bytes
ReadFile(hParseFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("Parse Index: Read did not read 16 bytes", true);
	return false;
	}
//
Index.Clip = clip;
//
Index.StartAddr = BufferToAddress(&buffer16[0]);    // bytes 0-2
Index.SA_LOBit = (buffer16[2] & 0x01);
//
Index.EndRepAddr = BufferToAddress(&buffer16[3]);   // bytes 3-5
Index.ERA_LOBit = (buffer16[5] & 0x01);
//
Index.StartRepAddr = BufferToAddress(&buffer16[6]); // bytes 6-8
Index.SRA_LOBit = (buffer16[8] & 0x01);
//
Index.SMInit = (buffer16[9] << 16) | (buffer16[10] << 8) | buffer16[11];    // bytes 9-11
//
Index.EndAddr = BufferToAddress(&buffer16[12]);     // bytes 12-14
Index.EA_LOBit = (buffer16[14] & 0x01);
//
Index.databyte = buffer16[15];                      // byte 15
//
Index.ClipLength = (Index.EndAddr - Index.StartAddr) + 1;
Index.ClipRepLength = (Index.EndRepAddr - Index.StartRepAddr) + 1;
//
return true;
}
//---------------------------------------------------------------------------

bool WriteIndex(SFIndex &Index, HANDLE hParseFile, int clip)
{
// write a 16 byte index entry from the SFIndex structure named Index
// clip number is 1 based
SetFilePointer(hParseFile, 0x0100 + ((clip - 1) * 16), NULL, FILE_BEGIN);
unsigned char buffer16[16];
AddressToBuffer(Index.StartAddr, &buffer16[0]);
buffer16[2] |= Index.SA_LOBit;
//
AddressToBuffer(Index.EndRepAddr, &buffer16[3]);
buffer16[5] |= Index.ERA_LOBit;
//
AddressToBuffer(Index.StartRepAddr, &buffer16[6]);
buffer16[8] |= Index.SRA_LOBit;
//
buffer16[9] = (Index.SMInit >> 16) & 0xFF;
buffer16[10] = (Index.SMInit >> 8) & 0xFF;
buffer16[11] = (Index.SMInit) & 0xFF;
//
AddressToBuffer(Index.EndAddr, &buffer16[12]);
buffer16[14] |= Index.EA_LOBit;
//
buffer16[15] = Index.databyte;
//
DWORD byteswritten;
WriteFile(hParseFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("Write Index: Write did not write 16 bytes", true);
	return false;
	}
return true;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::AnalyzeButtonClick(TObject *Sender)
{
HANDLE hAnalyzeFile;/* file handle */

// Can only analyze mth files
String Extension = LowerCase(ExtractFileExt(MRUComboBox->Text));
if (Extension != ".mth") {
	MyShowMessage("Analyze: Can only analyze '.mth' files", true);
	return;
	}
/*
 * try and open the file
 */
if((hAnalyzeFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Analyze: Unable to open file: " + MRUComboBox->Text, true);
	return;
	} else {
	MyShowMessage("Analyze: Analyzing file: " + MRUComboBox->Text, false);
	}

/*
A&S SW1200.mth
		  ---1---  ---2---  ---3---  ---4---  ---5---
00000100  01 00 00 FF FF FF FF FF FF 00 00 00 02 05 A7 0B  .........
00000110  02 05 A8 FF FF FF FF FF FF 00 00 00 03 3C 7D 0B  ......<}.
00000120  FF FF FF FF FF FF FF FF FF 00 00 00 FF FF FF 00  ....
....
000001A0  FF FF FF FF FF FF FF FF FF 00 00 00 FF FF FF 00  ....
000001B0  04 29 16 FF FF FF FF FF FF 00 00 00 04 69 55 1B  .).....iU.
000001C0  04 69 56 FF FF FF FF FF FF 00 00 00 04 BC 39 1B  .iV....9.
000001D0  04 BC 3A FF FF FF FF FF FF 00 00 00 05 04 79 1B  .:.....y.
000001E0  FF FF FF FF FF FF FF FF FF 00 00 00 FF FF FF 00  ....
000001F0  FF FF FF FF FF FF FF FF FF 00 00 00 FF FF FF 00  ....
...
000003B0  08 0F 1E FF FF FF FF FF FF 00 00 00 08 50 EB 2B  .......P+
000003C0  FF FF FF FF FF FF FF FF FF 00 00 00 FF FF FF 00  ....
000003D0  08 50 EC 08 9D 34 08 50 EC 00 00 00 08 9D 35 1F  .P..4.P.....5.
000003E0  FF FF FF FF FF FF FF FF FF 00 00 00 FF FF FF 00  ....
*/

getFileDetails(hAnalyzeFile);

// empty String Grid
ClearStringGrid();
//
PSGroup->ItemIndex = PS2;	// default group
if (SoundFileSize > 0x200000) {
	PSGroup->ItemIndex = PS3;	// if memory is greater than 2MB, its definitely PS3
	} else {
	PSGroup->ItemIndex = PS2;	// PS2 but still could be PS3 if 24 bit PCM clips are found
	}

// Fill String Grid
SFIndex LocalIndex;
EndAddrMax = 0;
for (int indexclipNo = 1 ; indexclipNo <= 255 ; indexclipNo++) {	// goes through the index : 1 based

	//---------------------------------------------------------------------------
	// zeorth Column of data is the Clip Number or the offset in Hex
	StringGrid1->Cells[ICLIPNO][indexclipNo] = HEXFIELDPREFIX + IntToHex(0x0100 + ((indexclipNo - 1) * 16), IOFFSETDIGITSTODISPLAY);
	// read and decode 16 bytes
	if (!ParseIndex(LocalIndex, hAnalyzeFile, indexclipNo)) {
		break;	// error parsing index
		}
	if (LocalIndex.StartAddr == EMPTY_FIELD) continue;	// on to the next index value
	//---------------------------------------------------------------------------
	// Seventh column of data - one byte
	StringGrid1->Cells[IDATABYTE][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalIndex.databyte, 2);
	//
	if (AutoCheckBox->Checked && (LocalIndex.databyte & PCM24MASK)) {
		PSGroup->ItemIndex = PS3;	//set PS3 Format - but not always true for PS3 files
									// later, set EEPROM size to 4MB.
		}
	int iflag = (int) StringGrid1->Objects[IDATABYTE][indexclipNo];
	iflag = (iflag & 0xFFFFFF00) | LocalIndex.databyte;
	iflag = (iflag & 0xFFFF00FF) | (LocalIndex.SA_LOBit << 11) | (LocalIndex.ERA_LOBit << 10) | (LocalIndex.SRA_LOBit << 9) | (LocalIndex.EA_LOBit << 8);
	StringGrid1->Objects[IDATABYTE][indexclipNo] = (TObject *) iflag;
	//---------------------------------------------------------------------------
	// First column of data - Starting address of clip
	StringGrid1->Cells[ISTARTADDR][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalIndex.StartAddr, IDIGITSTODISPLAY);

	if (indexclipNo == 1) DebugBuffer->Lines->Append("Analyze: Start of Sounds " +  StringGrid1->Cells[ISTARTADDR][indexclipNo]);
	//---------------------------------------------------------------------------
	// Second column of data - End address of repeating segment
	if (LocalIndex.EndRepAddr != EMPTY_FIELD) {
		StringGrid1->Cells[IENDREPADDR][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalIndex.EndRepAddr, IDIGITSTODISPLAY);
		}
	//---------------------------------------------------------------------------
    // Third column of data - Start address of repeating segment
	if (LocalIndex.StartRepAddr != EMPTY_FIELD) {
		StringGrid1->Cells[ISTARTREPADDR][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalIndex.StartRepAddr, IDIGITSTODISPLAY);
		if (LocalIndex.StartRepAddr > LocalIndex.EndRepAddr) {	// check values
			//MyShowMessage("Analyze: Repeating Clip Addreses incorrect at Offset " + StringGrid1->Cells[ICLIPNO][indexclipNo], true);
			MyShowMessage("Analyze: Repeating Clip Addreses incorrect for clip #" + String(indexclipNo), true);
			}
    	}
	//---------------------------------------------------------------------------
	// Fourth column of data - IMA ADPCM State Machine Initialization and if used for PCM24 unknown meaning
	StringGrid1->Cells[ISMINIT][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalIndex.SMInit,6);
	//---------------------------------------------------------------------------
	// Fifth column of data - Ending address of clip
	if (LocalIndex.EndAddr != EMPTY_FIELD) {
		//int EndAddr = BufferToAddress(&buffer16[12]);
		StringGrid1->Cells[IENDADDR][indexclipNo] = HEXFIELDPREFIX + IntToHex(LocalIndex.EndAddr, IDIGITSTODISPLAY);
		// find highest end address
		if (LocalIndex.EndAddr > EndAddrMax) EndAddrMax = LocalIndex.EndAddr;

    	}
	//---------------------------------------------------------------------------
	// Sixth column of data - calculated length of clip
	//LocalIndex.ClipLength = (LocalIndex.EndAddr - LocalIndex.StartAddr) + 1;
	StringGrid1->Cells[ILENGTH][indexclipNo] = LocalIndex.ClipLength;
	//---------------------------------------------------------------------------
	}

MyShowMessage("Analyze: End of Sounds 0x" + IntToHex(EndAddrMax, IDIGITSTODISPLAY), false);
CloseHandle(hAnalyzeFile);

// check the low RAM checksum
(void) ChecksumCheck(MRUComboBox->Text);

Decoded = true; 	// signal sound file analyzed
ExportOneCliptomp3->Enabled = true;
EditIndexButton->Enabled = true;
EditClipButton->Enabled = true;
EditSoftkeysButton->Enabled = true;
EditParamsButton->Enabled = true;
// display Clip # numbers
ToggleOffsetClipNoClick(Sender);
// set push buttons
StringGrid1Click(Sender);
AvailableTimeRemainingClick(Sender);
// focus on the StringGrid
StringGrid1->SetFocus();
// make the selected clip visible
StringGrid1->TopRow = (StringGrid1->Row - 3) >= 1 ? (StringGrid1->Row - 3) : 1;
return;
}
//---------------------------------------------------------------------------
bool __fastcall TADPCM_Main::ChecksumCheck(String FileName)
{
bool bretv = false;
WORD cksum = 0, value;
// open the file
HANDLE hSoundFile;
// try and open the file
if((hSoundFile = CreateFile(FileName.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	ADPCM_Main->MyShowLastError("CHecksum: Unable to open file : " + FileName, true);
	Close();    // cancel the operation
	} else {
	if (iDebug > 0) ADPCM_Main->MyShowMessage("Checksum: Sound file to be checked : " + FileName, false);
	}
// sum up the words from 0x00 to 0x65 - result should be 0xFFFF
// 51 words

for (int idx = 0; idx <= 50 ; idx++) {
	SetFilePointer(hSoundFile, idx * 2, NULL, FILE_BEGIN);

	unsigned char buffer16[2];
	DWORD readBytes;
	// read 2 bytes
	ReadFile(hSoundFile, buffer16, sizeof(buffer16), &readBytes, NULL);
	if (readBytes != 2) {
		MyShowMessage("Checksum: Read did not read 2 bytes", true);
		bretv = false;
		}
	value = ((buffer16[0] << 8) | buffer16[1]);
	cksum += value;
	//MyShowMessage("Checksum index " + String(idx * 2) + " : " + IntToHex( cksum, 4), false);
	}
if (cksum == 0xFFFF) {
	MyShowMessage("Checksum: OK (0x" + IntToHex(value, 4) + ")", false);
	bretv = true;
	} else {
	MyShowMessage("Checksum: Error (0x" + IntToHex(value, 4) + ")", false);
	bretv = false;
	}
CloseHandle(hSoundFile);
return bretv;	
}

//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::BrowseforFileButtonClick(TObject *Sender)
{
OpenDialog->DefaultExt = "mth";
OpenDialog->Filter = "Sound files (*.mth)|*.MTH|Audio Clip file (*.ima *.raw)|*.ima;*.IMA;*.raw;*.RAW";
OpenDialog->Options.Clear();
OpenDialog->Options << ofFileMustExist;
OpenDialog->Title = "Select Sound File to Analyze or Play";

if (OpenDialog->Execute()) {
	// file picked
	MRUComboBox->Text = OpenDialog->FileName;
	MRUComboBox->Hint = MRUComboBox->Text;
	ClearButtonClick(Sender);
	EngineEEPROMSize->Caption = "";
	Decoded = false;
	OpenDialog->InitialDir = ExtractFilePath(OpenDialog->FileName);

    // insert into MRU List
    // don't insert a duplicate string
    int idx = MRUComboBox->Items->IndexOf(MRUComboBox->Text);
    if (idx == -1) {
        MRUComboBox->Items->Insert(0, MRUComboBox->Text);
        MRUComboBox->ItemIndex = 0;
		if (MRUComboBox->Items->Count > MRUComboBox->DropDownCount) MRUComboBox->Items->Delete(MRUComboBox->DropDownCount);
        } else {
        // set ItemIndex
        MRUComboBox->ItemIndex = idx;
		}
	// normal default values
	PSGroup->ItemIndex = PS2;
	FormatGroup->ItemIndex = ADPCM;
	SampleGroup->ItemIndex = HZ5512;
	}
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StopButton2Click(TObject *Sender)
{
Stopem = true;
(void) waveOutReset(hWaveOut);
MyShowMessage("Playback is Stopping", false);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::PlayClipButtonClick(TObject *Sender)
{
// play the entire clip
Overflow->Visible = false;
Overflow->Checked = false;
//
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}

if (StringGrid1->Cells[ISTARTADDR][StringGrid1->Row] == "") {
	MyShowMessage("Play Clip: Empty clip - try again", true);
	return;
	}

// (Start Address, End Address, index, predValue)
short index = 0;
short predValue = 0;
//
// This is for a case seen in some PS3 file. The State Machine Initialization values
// are used even for non-repeating clips.
// TODO This needs to be thought about more.
//
// Use _tcstol for both BCB4 and CB10          see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
if (_tcstol(StringGrid1->Cells[ISTARTREPADDR][StringGrid1->Row].c_str(), NULL, 16) == 0) {  // non-repeating clip?
	// pick out the IMA ADPCM state machine's index and previous value
	index = _tcstol(StringGrid1->Cells[ISMINIT][StringGrid1->Row].c_str(), NULL, 16) >> 16;
	predValue = _tcstol(StringGrid1->Cells[ISMINIT][StringGrid1->Row].c_str(), NULL, 16) & 0x0000FFFF;
	if (index < -1 || index > 88) {
		MyShowMessage("Play Clip: IMA ADPCM index out of range: " + String(index), false);
		return;
		}
	}

TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

// Use _tcstol for both BCB4 and CB10          see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][StringGrid1->Row].c_str(), NULL, 16);
const int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][StringGrid1->Row].c_str(), NULL, 16);
const int StartRepAddr = _tcstol(StringGrid1->Cells[ISTARTREPADDR][StringGrid1->Row].c_str(), NULL, 16);
PlayClip(StartAddr, 0, 0, EndAddr, StartRepAddr, index, predValue);
Screen->Cursor = OldCursor;
return;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::PauseButtonClick(TObject *Sender)
{
(void) waveOutPause(hWaveOut);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::RestartButtonClick(TObject *Sender)
{
(void) waveOutRestart(hWaveOut);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1DblClick(TObject *Sender)
{
TPoint pt;
GetCursorPos(&pt);  // current mouse position
int Column, Row;
pt = StringGrid1->ScreenToClient(pt);
StringGrid1->MouseToCell(pt.x, pt.y, Column, Row);
if (Row == 0) return; // ignore double clicks in fixed rows and columns
if (Column == 0) return;

if (PlayClipButton->Enabled) PlayClipButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::GetDevCapsButtonClick(TObject *Sender)
{
// from:
// http://stackoverflow.com/questions/25633019/why-does-waveoutopen-on-windows-vista-newer-say-the-default-output-device-can
//
MMRESULT err;
WAVEOUTCAPS caps;

ZeroMemory(&caps, sizeof (WAVEOUTCAPS));
err = waveOutGetDevCaps(WAVE_MAPPER, &caps, sizeof (WAVEOUTCAPS));
if (err != MMSYSERR_NOERROR) {
	DebugBuffer->Lines->Append("GetDevCaps: mmsys err " + String(err));
	return;
	}
	DebugBuffer->Lines->Append("Results of waveOutGetDevCaps()  ----------------");
    DebugBuffer->Lines->Append("Name: " + String(caps.szPname));
	DebugBuffer->Lines->Append("Version: " + IntToHex((int)caps.vDriverVersion, 4));
    DebugBuffer->Lines->Append("formats: ");
#define FORMATS(x) if((caps.dwFormats & x) != 0) DebugBuffer->Lines->Append(#x);
    FORMATS(WAVE_FORMAT_1M08);	// XYdd : X = 11.025, 22.05, 44.1 kHz
    FORMATS(WAVE_FORMAT_1M16);	//        Y = Stereo/Mono
    FORMATS(WAVE_FORMAT_1S08);	//        dd = 8/16 bit
    FORMATS(WAVE_FORMAT_1S16);
    FORMATS(WAVE_FORMAT_2M08);
    FORMATS(WAVE_FORMAT_2M16);
    FORMATS(WAVE_FORMAT_2S08);
    FORMATS(WAVE_FORMAT_2S16);
    FORMATS(WAVE_FORMAT_4M08);
    FORMATS(WAVE_FORMAT_4M16);
    FORMATS(WAVE_FORMAT_4S08);
    FORMATS(WAVE_FORMAT_4S16);
#define WAVE_FORMAT_44M08      0x00000100       /* 44.1   kHz, Mono,   8-bit  */
#define WAVE_FORMAT_44S08      0x00000200       /* 44.1   kHz, Stereo, 8-bit  */
#define WAVE_FORMAT_44M16      0x00000400       /* 44.1   kHz, Mono,   16-bit */
#define WAVE_FORMAT_44S16      0x00000800       /* 44.1   kHz, Stereo, 16-bit */
#define WAVE_FORMAT_48M08      0x00001000       /* 48     kHz, Mono,   8-bit  */
#define WAVE_FORMAT_48S08      0x00002000       /* 48     kHz, Stereo, 8-bit  */
#define WAVE_FORMAT_48M16      0x00004000       /* 48     kHz, Mono,   16-bit */
#define WAVE_FORMAT_48S16      0x00008000       /* 48     kHz, Stereo, 16-bit */
#define WAVE_FORMAT_96M08      0x00010000       /* 96     kHz, Mono,   8-bit  */
#define WAVE_FORMAT_96S08      0x00020000       /* 96     kHz, Stereo, 8-bit  */
#define WAVE_FORMAT_96M16      0x00040000       /* 96     kHz, Mono,   16-bit */
#define WAVE_FORMAT_96S16      0x00080000       /* 96     kHz, Stereo, 16-bit */
    FORMATS(WAVE_FORMAT_44M08);       /* 44.1   kHz, Mono,   8-bit  */
    FORMATS(WAVE_FORMAT_44S08);       /* 44.1   kHz, Stereo, 8-bit  */
    FORMATS(WAVE_FORMAT_44M16);       /* 44.1   kHz, Mono,   16-bit */
    FORMATS(WAVE_FORMAT_44S16);       /* 44.1   kHz, Stereo, 16-bit */
    FORMATS(WAVE_FORMAT_48M08);       /* 48     kHz, Mono,   8-bit  */
    FORMATS(WAVE_FORMAT_48S08);       /* 48     kHz, Stereo, 8-bit  */
    FORMATS(WAVE_FORMAT_48M16);       /* 48     kHz, Mono,   16-bit */
    FORMATS(WAVE_FORMAT_48S16);       /* 48     kHz, Stereo, 16-bit */
    FORMATS(WAVE_FORMAT_96M08);
    FORMATS(WAVE_FORMAT_96M16);
	FORMATS(WAVE_FORMAT_96S08);
    FORMATS(WAVE_FORMAT_96S16);

	DebugBuffer->Lines->Append("channels: " + String(caps.wChannels));

    DebugBuffer->Lines->Append("supports: ");
#define SUPPORTS(x) if((caps.dwSupport & x) != 0) DebugBuffer->Lines->Append(#x);
    SUPPORTS(WAVECAPS_LRVOLUME);		/* separate left-right volume control */
    SUPPORTS(WAVECAPS_PITCH);			/* supports pitch control */
    SUPPORTS(WAVECAPS_PLAYBACKRATE);
    SUPPORTS(WAVECAPS_SYNC);
    SUPPORTS(WAVECAPS_VOLUME);
    SUPPORTS(WAVECAPS_SAMPLEACCURATE);
#define WAVECAPS_DIRECTSOUND 0x0040
	SUPPORTS(WAVECAPS_DIRECTSOUND);

}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ExitButtonClick(TObject *Sender)
{
Close();	
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1Click(TObject *Sender)
{
if (StopButton1->Enabled) return;	// do nothing if playing a clip

if (StringGrid1->Cells[ISTARTADDR][StringGrid1->Row] == "") {
	PlayClipButton->Enabled = false;
	EditIndexButton->Enabled = true;
	EditClipButton->Enabled = true;
	} else {
	PlayClipButton->Enabled = true;
	EditIndexButton->Enabled = true;
	EditClipButton->Enabled = true;
	}

// set Repeating Segment
if (StringGrid1->Cells[IENDREPADDR][StringGrid1->Row] == "") {
	PlayRepeatingSegmentButton->Enabled = false;
	LongPlayButton->Enabled = false;
    } else {
    PlayRepeatingSegmentButton->Enabled = true;
	LongPlayButton->Enabled = true;
	}

// set Sample Rate and Data Format
AutoCheckBoxClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::PlayRepeatingSegmentButtonClick(TObject *Sender)
{
// play repeating segment of clip
Overflow->Visible = false;
Overflow->Checked = false;
// pick out the IMA ADPCM state machine's index and previous value
// Use _tcstol for both BCB4 and CB10         see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const short index = _tcstol(StringGrid1->Cells[ISMINIT][StringGrid1->Row].c_str(), NULL, 16) >> 16;
const short predValue = _tcstol(StringGrid1->Cells[ISMINIT][StringGrid1->Row].c_str(), NULL, 16) & 0x0000FFFF;

if (index < -1 || index > 88) {
	MyShowMessage("Play Repeating Segment: IMA ADPCM index out of range: " + String(index), false);
	return;
	}
TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
// Use _tcstol for both BCB4 and CB10         see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][StringGrid1->Row].c_str(), NULL, 16);
int StartRepAddr = _tcstol(StringGrid1->Cells[ISTARTREPADDR][StringGrid1->Row].c_str(), NULL, 16);
const int EndRepAddr = _tcstol(StringGrid1->Cells[IENDREPADDR][StringGrid1->Row].c_str(), NULL, 16);

if (FormatGroup->ItemIndex == PCM24) {   // 24 pit PCM
	//StartRepAddr = StartRepAddr - abs((StartRepAddr % 3) - (StartAddr % 3));
	if ((StartRepAddr % 3) != (StartAddr % 3)) StartRepAddr++;
	if ((StartRepAddr % 3) != (StartAddr % 3)) StartRepAddr++;
	}

// (Start Address, Start Repeat Address, End Repeat Address, End Address, state.step, state.predValue)
PlayClip(0, StartRepAddr, EndRepAddr, -1, 0, index, predValue);
Screen->Cursor = OldCursor;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::CopyRowtoClipboardClick(TObject *Sender)
{
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
//Copy the selected row of the clip index to the clipboard
ClipEdit->Text = StringGrid1->Rows[StringGrid1->Row]->CommaText;
ClipEdit->SelectAll();
ClipEdit->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::CopyClipIndextoClipboard1Click(
      TObject *Sender)
{
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
// Copy the entire clip index to the clipboard
MemoBuffer->Visible = false;
MemoBuffer->Clear();
MemoBuffer->Lines->Append("\"" + MRUComboBox->Text + "\"");
for (int i = 0 ; i <= 255  ; i++) {
	MemoBuffer->Lines->Append(StringGrid1->Rows[i]->CommaText);
	}
MemoBuffer->SelectAll();
MemoBuffer->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::VolumeUpDownClick(TObject *Sender,
	  TUDBtnType Button)
{
const short Pos = (VolumeUpDown->Position & 0x0F) << 12;
const DWORD Vol =  (Pos << 16) | Pos;
if (waveOutSetVolume(hWaveOut, Vol) != MMSYSERR_NOERROR)  {
	MyShowMessage("VolumeUpDown: waveOutSetVolume Failure", true);
    }
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::AutoCheckBoxClick(TObject *Sender)
{

if (!Decoded) {
	return;
	}

const int DataVal = _tcstol(StringGrid1->Cells[IDATABYTE][StringGrid1->Row].c_str(), NULL, 16);
// Use _tcstol  for both BCB4 and CB10           see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
if (AutoCheckBox->Checked && StringGrid1->Cells[ISTARTADDR][StringGrid1->Row] != "") {
	if (DataVal & PCM24MASK)
		FormatGroup->ItemIndex = PCM24;
		else
		FormatGroup->ItemIndex = ADPCM;

	// set the Sample Rate based on DataVal - this does our best guess
	int Sample = -1;
	if (PSGroup->ItemIndex == PS2) {
		// PS2
		if (DataVal & PCM24MASK) MyShowMessage("24 bit PCM clip in a PS2 sound file", true);
		//Sample = SampleRatePS2[DataVal & 0x7F]; // index 0-127
		// ADPCM
		if ((DataVal & 0x48) == 0x48) Sample = HZ5512;
		if ((DataVal & 0x48) == 0x08) Sample = HZ11025;
		if ((DataVal & 0x48) == 0x00) Sample = HZ22050;

		if (Sample == -1) {
			MyShowMessage("Can't decode the PS2 Sample Rate for Data Byte (defaults to 5512 Hz): " + StringGrid1->Cells[IDATABYTE][StringGrid1->Row], true);
			}
		} else {
		// PS3
		//Sample = SampleRatePS3[DataVal & 0xFF]; // index 0-255
		// ADPCM
		if ((DataVal & 0xC8) == 0x48) Sample = HZ5512;
		if ((DataVal & 0xC8) == 0x08) Sample = HZ11025;
		if ((DataVal & 0xC8) == 0x00) Sample = HZ22050;
		// PCM
		if ((DataVal & 0xC8) == 0x88) Sample = HZ5512;
		if ((DataVal & 0xC8) == 0x80) Sample = HZ11025;
		if (Sample == -1) {
			MyShowMessage("Can't decode the PS3 Sample Rate for Data Byte (defaults to 5512 Hz): " + StringGrid1->Cells[IDATABYTE][StringGrid1->Row], true);
			}
		}

	if (Sample != -1) {
		SampleGroup->ItemIndex = Sample;
		} else {
		SampleGroup->ItemIndex = HZ5512;    // can't decode the sample rate, default to 5512 Hz
		}
	switch((DataVal & 0x30) >> 4) {
		case EngineSound:
			SoundCategory->Caption = "Engine Sound Clip";
			break;
		case AccentSound:
			SoundCategory->Caption = "Accent Sound Clip";
			break;
		case HornWhistleSound:
			SoundCategory->Caption = "Horn/Whistle Sound Clip";
			break;
		case BellSound:
			SoundCategory->Caption = "Bell Sound Clip";
			break;
		}
	} else {
	SoundCategory->Caption = "";
	}
}
//---------------------------------------------------------------------------
void __fastcall TADPCM_Main::ClearStringGrid(void)
{
Decoded = false;
for(int x = StringGrid1->FixedRows; x < StringGrid1->RowCount; ++x)
	StringGrid1->Rows[x]->Clear();
// clear the Quill String Grid if created
if (Quill_Index!= NULL) Quill_Index->ClearStringGrid();
StringGrid1->Repaint();
}

//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ClearButtonClick(TObject *Sender)
{
ExportOneCliptomp3->Enabled = false;
PlayClipButton->Enabled = false;
LongPlayButton->Enabled = false;
PlayRepeatingSegmentButton->Enabled = false;
EditIndexButton->Enabled = false;
EditClipButton->Enabled = false;
EditSoftkeysButton->Enabled = false;
EditParamsButton->Enabled = false;

DebugBuffer->Clear();
EngineEEPROMSize->Caption = "";
ClearStringGrid();	// sets Decoded = false;
PSGroup->ItemIndex = PS2;

iDebug = 0;
Hidden->Hint = "!";
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::PlayEntireFileButtonClick(TObject *Sender)
{
// This plays the entire file using the format and data rate set in
// the radio buttons. Note that the format and data rate changes over
// different clips in the file so this routine makes some mightly awful
// noises
int StartAddr;
if (LowerCase(ExtractFileExt(MRUComboBox->Text)) == ".mth") {
	// must have analyzed a sound file
	if (!Decoded) {
		MyShowMessage("Press the Analyze button first", true);
		return;
		}
	Stopem = false;
	for (int i = 1 ; i <= 255  ; i++) {
		// set index into StringGrid
		StringGrid1->Row = i;   // if Automatic Settings is checked, this will set Clip Format and Sample Rate
		const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][i].c_str(), NULL, 16);    // Start
		if (StartAddr == 0) continue;       // skip empty index locations
		const int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][i].c_str(), NULL, 16);    // End
		PlayClip(StartAddr, 0, 0, EndAddr, 0, 0, 0);
        if (Stopem) break; // stop play
		}
	// play each clip in the index
	} else {
	// headerless files (.ima or .raw)
	// you must manually set the Clip Format and Sample Rate
	PlayClip(0x0000, 0, 0, -1, 0, 0, 0);    // play to end of file
	}
// MyShowMessage("Play Entire File: Complete", true);
return;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1KeyPress(TObject *Sender,
      char &Key)
{
if (Key == VK_RETURN) {
	StringGrid1DblClick(Sender);
    }	
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::MessagesCopytoClipboardClick(TObject *Sender)
{
DebugBuffer->SelectAll();
DebugBuffer->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::AboutADPCMClick(TObject *Sender)
{
// Splash Window Information Test
Splash = new TSplash(Application);
Splash->Memo1->Lines->Strings[0] = "ADPCM - " + String(VersionLabel->Caption);
Splash->Show();			// Show Splash Window
Splash->Update();		// Paint Splash Window
SS = SPLASH_OFF; 		// Turn off future Splash Window
AboutADPCM->Checked = true;
AboutADPCM1->Checked = true;
AboutADPCM2->Checked = true;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ADPCMWebPage1Click(TObject *Sender)
{
ShellExecute(NULL, TEXT("open"), TEXT("http://www.silogic.com/trains/ADPCM.html"), NULL, NULL, SW_RESTORE);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::FormCreate(TObject *Sender)
{

// read Preferences from registry if they exist
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);

//MRUComboBox->Text = RegFile->ReadString("Options","FileName", MRUComboBox->Text);
MRUComboBox->Hint = MRUComboBox->Text;
VolumeUpDown->Position = (short) RegFile->ReadInteger("Options","VolumeUpDown", VolumeUpDown->Position);
VolumeUpDownClick(NULL, (TUDBtnType) NULL);
SampleGroup->ItemIndex = RegFile->ReadInteger("Options","SampleRate", SampleGroup->ItemIndex);
FormatGroup->ItemIndex = RegFile->ReadInteger("Options","FormatGroup", FormatGroup->ItemIndex);
PSGroup->ItemIndex = RegFile->ReadInteger("Options","PSGroup", PSGroup->ItemIndex);
getcwd(cwd, MAXDIR);
AutoCheckBox->Checked = RegFile->ReadInteger("Options","AutoCheckBox", AutoCheckBox->Checked);
SS = RegFile->ReadInteger("Options", "SS", SS);	// this gets set to SPLASH_OFF when you click on "ADPCM Help"
if (SS == SPLASH_OFF) {
	AboutADPCM->Checked = true;
	AboutADPCM1->Checked = true;
   	AboutADPCM2->Checked = true;
    }
NoSoundFileSizeChecks->Checked = RegFile->ReadInteger("Options", "NoSoundFileSizeChecks", false);
if (NoSoundFileSizeChecks->Checked) {
	EngineEEPROMSize->Font->Color = clRed;
    }
RepUpDown->Position = (short) RegFile->ReadInteger("Options","RepUpDown", RepUpDown->Position);
StringGrid1->Row = RegFile->ReadInteger("Options", "ClipNo", StringGrid1->Row);

//http://bcbjournal.org/articles/vol1/9711/Implementing_a_recent-files_list.htm
for (int idx = 0 ; idx < MRUComboBox->DropDownCount ; idx++) {
	String S = RegFile->ReadString("MRU Sound File", "MRU" + String(idx + 1), "");
	if (S != "") {
		MRUComboBox->Items->Append(S);
        } else {
        break;
        }
    }
// next line changes MRUComboBox->Text
MRUComboBox->ItemIndex = RegFile->ReadInteger("MRU Sound File", "MRUItemIndex", -1);
MRUComboBox->Hint = MRUComboBox->Text;;
// Deallocate RegFile
delete RegFile;
//
//
if (MRUComboBox->Text.IsEmpty()) {
    // use WorkingDir
	OpenDialog->InitialDir = cwd;
	MRUComboBox->Text = OpenDialog->InitialDir + "\\*.mth";
   	MRUComboBox->Hint = MRUComboBox->Text;
    } else {
    // use directory from Script FileName
	OpenDialog->InitialDir = ExtractFilePath(MRUComboBox->Text);
   	MRUComboBox->Hint = MRUComboBox->Text;
	}

if (SS == SPLASH_ON) {
	// Splash Window
    Splash = new TSplash(Application);
	Splash->Memo1->Lines->Strings[0] = "ADPCM - " + String(VersionLabel->Caption);
    Splash->Show();			// Show Splash Window
    Splash->Update();		// Paint Splash Window
	}

DragAcceptFiles (Handle, True);		//  Accept drag and drop from the Explorer

ProgramName = ParamStr(0);
iDebug = 0;
String Parameter;
// Check for command line files
const int NumDrop = ParamCount();
if (NumDrop > 0) {
	for (int i = 1; i <= NumDrop; i++) {
       	Application->ProcessMessages();
		Parameter = ParamStr(i);
		String str = "ParamStr(" + String(i) + ") ";
		//
		if (iDebug > 0) Application->MessageBox(ParamStr(i).c_str(), str.c_str(), MB_OK | MB_DEFBUTTON1);
		MyShowMessage(str + ParamStr(i), false);
		//
		// See if its a parameter starting with /
        if (Parameter[1] == '/') {
			// program does not use parameter yet for anything other than debug
			if (iDebug > 0) Application->MessageBox(Parameter.c_str(), TEXT("Parameter"), MB_OK | MB_DEFBUTTON1);
			MyShowMessage("Parameter: " + Parameter, false);
			if (Parameter[2] == 'd') {
				iDebug++;
                }
            //
            continue;
            }

		if (FileExists(Parameter.c_str())) {
			// its a filename
			if (iDebug > 0) Application->MessageBox(Parameter.c_str(), TEXT("File Exists"), MB_OK | MB_DEFBUTTON1);
			//MyShowMessage("File Exists", false);
			MRUComboBox->Text = Parameter;
			MRUComboBox->Hint = MRUComboBox->Text;
			OpenDialog->InitialDir = ExtractFilePath(MRUComboBox->Text);
			Update();	// redraw Form
			} else {
			if (DirectoryExists(Parameter.c_str())){
				// its a directory not a filename
				if (iDebug > 0) Application->MessageBox(Parameter.c_str(), TEXT("Directory Exists"), MB_OK | MB_DEFBUTTON1);
				//MyShowMessage("Directory Exists", false);
				OpenDialog->InitialDir = Parameter;
				break; // Stop after first directory because it does not
					// make sense to have multiple directories on the command line
				} else {
				MyShowMessage("File or Directory not found", false);
				}
        	}

		// use last dropped path/filename
		MRUComboBox->Text = Parameter;
		MRUComboBox->Hint = MRUComboBox->Text;
		OpenDialog->InitialDir = ExtractFilePath(MRUComboBox->Text);

		// Add new filename to DropDown List
		// insert into MRU List
		// don't insert a duplicate string
		int idx = MRUComboBox->Items->IndexOf(MRUComboBox->Text);
		if (idx == -1) {
			MRUComboBox->Items->Insert(0, MRUComboBox->Text);
			MRUComboBox->ItemIndex = 0;
			if (MRUComboBox->Items->Count > MRUComboBox->DropDownCount)
			MRUComboBox->Items->Delete(MRUComboBox->DropDownCount);
			} else {
			// set ItemIndex
			MRUComboBox->ItemIndex = idx;
			}
		// set Clip Format and Sample Rate
		if (MRUComboBox->Text.Pos(".ima")) FormatGroup->ItemIndex = ADPCM; 
		if (MRUComboBox->Text.Pos(".raw")) FormatGroup->ItemIndex = PCM24; 
		if (MRUComboBox->Text.Pos("5512")) SampleGroup->ItemIndex = HZ5512; 
		if (MRUComboBox->Text.Pos("8000")) SampleGroup->ItemIndex = HZ8000;
		if (MRUComboBox->Text.Pos("11025")) SampleGroup->ItemIndex = HZ11025;
		if (MRUComboBox->Text.Pos("16000")) SampleGroup->ItemIndex = HZ16000;
		if (MRUComboBox->Text.Pos("22050")) SampleGroup->ItemIndex = HZ22050;
		if (MRUComboBox->Text.Pos("44100")) SampleGroup->ItemIndex = HZ44100;
		}
	}
return;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::FormDestroy(TObject *Sender)
{
//
// write preferences to the Registry
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);
//RegFile->WriteString("Options", "FileName", MRUComboBox->Text);
RegFile->WriteInteger("Options", "VolumeUpDown", VolumeUpDown->Position);
RegFile->WriteInteger("Options", "SampleRate", SampleGroup->ItemIndex);
RegFile->WriteInteger("Options", "FormatGroup", FormatGroup->ItemIndex);
RegFile->WriteInteger("Options", "PSGroup", PSGroup->ItemIndex);
RegFile->WriteInteger("Options", "AutoCheckBox", AutoCheckBox->Checked);
RegFile->WriteInteger("Options", "SS", SS);
RegFile->WriteInteger("Options", "NoSoundFileSizeChecks", NoSoundFileSizeChecks->Checked);
RegFile->WriteInteger("Options", "RepUpDown", RepUpDown->Position);
RegFile->WriteInteger("Options", "ClipNo", StringGrid1->Row);
// Copyright
RegFile->WriteString("Copyright", "Symbol", "");
RegFile->WriteString("Copyright", "By", "Mark C. DiVecchio");
RegFile->WriteString("Copyright", "Year", "2015");

// erase MRU PC key
RegFile->EraseSection("MRU Sound File");
// then resave the MRU List
int Cnt = MRUComboBox->Items->Count;
if (Cnt > 0) {
	for (int idx = 0 ; idx < Cnt ; idx++) {
		RegFile->WriteString("MRU Sound File", "MRU" + String(idx + 1), MRUComboBox->Items->Strings[idx]);
		}
	}
RegFile->WriteInteger("MRU Sound File", "MRUItemIndex", MRUComboBox->ItemIndex);
// Deallocate RegFile
delete RegFile;
return;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::WMDropFiles(TWMDropFiles &message)
{
// Get the number of files dropped
const UINT filecount = DragQueryFile((HDROP) message.Drop, 0xFFFFFFFF, NULL, 0);
if (filecount > 0) {
	ClearButtonClick(NULL);
	Decoded = false;
	String FullFileName;
	for (UINT ii = 0 ; ii < filecount ; ++ii){
    	Application->ProcessMessages();
		// Get the file name
		TCHAR FileName[MAX_PATH];
		DragQueryFile ((HDROP) message.Drop, ii, FileName, MAX_PATH);
		FullFileName = FileName;
		//Application->MessageBox(FullFileName.c_str(), TEXT("Result of DragQueryFile()"), MB_OK | MB_DEFBUTTON1);
		if (FileExists(FullFileName.c_str())) {
			// do something for each file
			MyShowMessage("Dropped file " + FullFileName, false);
			} else {
			continue;
			}
		}
	String FileName = LowerCase(ExtractFileName(FullFileName));
	// use last dropped path/filename
	MRUComboBox->Text = FullFileName;
	if (Quill_Index != NULL) Quill_Index->FileEdit->Text = FullFileName;
	MRUComboBox->Hint = MRUComboBox->Text;
	if (Quill_Index != NULL) Quill_Index->FileEdit->Hint = MRUComboBox->Hint;

	OpenDialog->InitialDir = ExtractFilePath(MRUComboBox->Text);

	// Add new filename to DropDown List
	// insert into MRU List
	// don't insert a duplicate string
	int idx = MRUComboBox->Items->IndexOf(MRUComboBox->Text);
	if (idx == -1) {
		MRUComboBox->Items->Insert(0, MRUComboBox->Text);
		MRUComboBox->ItemIndex = 0;
		if (MRUComboBox->Items->Count > MRUComboBox->DropDownCount)
		MRUComboBox->Items->Delete(MRUComboBox->DropDownCount);
		} else {
        // set ItemIndex
        MRUComboBox->ItemIndex = idx;
		}
	// set Clip Format and Sample Rate
	if (MRUComboBox->Text.Pos(".ima")) FormatGroup->ItemIndex = ADPCM; 
	if (MRUComboBox->Text.Pos(".raw")) FormatGroup->ItemIndex = PCM24; 
	if (MRUComboBox->Text.Pos("5512")) SampleGroup->ItemIndex = HZ5512; 
	if (MRUComboBox->Text.Pos("8000")) SampleGroup->ItemIndex = HZ8000;
	if (MRUComboBox->Text.Pos("11025")) SampleGroup->ItemIndex = HZ11025;
	if (MRUComboBox->Text.Pos("16000")) SampleGroup->ItemIndex = HZ16000;
	if (MRUComboBox->Text.Pos("22050")) SampleGroup->ItemIndex = HZ22050;
	if (MRUComboBox->Text.Pos("44100")) SampleGroup->ItemIndex = HZ44100;
	}

Update();	// redraw Form
DragFinish ((HDROP) message.Drop);
if (Edit_Engine_Parameters) {
	delete Edit_Engine_Parameters;
	Edit_Engine_Parameters = NULL;
	}
if (Edit_Scale_Factor) {
	delete Edit_Scale_Factor;
	Edit_Scale_Factor = NULL;
	}
	
return;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1FixedCellClick(TObject *Sender, int ACol,
		  int ARow)
{
// This toggles the values in the first column from memory addresses to/from
// numbers. The numbers correspond with the parameter to the 'n' command
// to the Engine.
if (ACol == 0 && ARow == 0) {
	ToggleOffsetClipNoClick(Sender);
	}
if (ARow != 0) {
	StringGrid1->Row = ARow;    // hightlight selected row
	}
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ToggleOffsetClipNoClick(TObject *Sender)
{
if (StringGrid1->Cells[ICLIPNO][1] == 1) { // if its numbers, change to addresses
	StringGrid1->Cells[ICLIPNO][0] = "Offset";
	for (int Row = 1 ; Row <= 255 ; Row++){
		StringGrid1->Cells[ICLIPNO][Row] = HEXFIELDPREFIX + IntToHex(0x0100 + ((Row - 1) * 16), IOFFSETDIGITSTODISPLAY);
		}
	SFS_Index->Caption = "SFS (0x3A0)";
	Bell_Index->Caption = "Bell (0xAC0)";
	Toot_Index->Caption = "Whistle/Horn (0xA70)";
	PFA_Index->Caption = "PFA (0x5F0)";
	Clickity_Clack_Index->Caption = "Clickity Clack (0x3D0)";
	EngineSounds_Index->Caption = "Engine Sounds (0xC00-0xC90)";
	SXS_Index->Caption = "SXS (0x10E0)";
	} else {
	// its addresses, change to numbers
	StringGrid1->Cells[ICLIPNO][0] = "Clip #";
	for (int Row = 1 ; Row <= 255 ; Row++){
		StringGrid1->Cells[ICLIPNO][Row] = Row;
		}
	SFS_Index->Caption = "SFS (43)";
	Bell_Index->Caption = "Bell (157)";
	Toot_Index->Caption = "Whistle/Horn (152)";
	PFA_Index->Caption = "PFA (80)";
	Clickity_Clack_Index->Caption = "Clickity Clack (46)";
	EngineSounds_Index->Caption = "Engine Sounds (177-186)";
	SXS_Index->Caption = "SXS (255)";
	}
IndexFormat = !IndexFormat;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1DrawCell(TObject *Sender, int ACol, int ARow,
		  TRect &Rect, TGridDrawState State)
{
// Default to black on white
StringGrid1->Canvas->Font->Color = clBlack;	// Normal Black on White
StringGrid1->Canvas->Brush->Color = clWhite;
// Draw the contents of each cell
if (Decoded) {
	StringGrid1->Canvas->Font->Style = StringGrid1->Canvas->Font->Style >> fsStrikeOut;
	// if the cell is selected, highlight it
	if( State.Contains(gdSelected) ) {
		StringGrid1->Canvas->Font->Color = clWhite;
		StringGrid1->Canvas->Brush->Color = clBlue;		// White on Blue
		} else {
		StringGrid1->Canvas->Font->Color = clBlack;	// Normal Black on White
		StringGrid1->Canvas->Brush->Color = clWhite;
		}
	} 

if (Decoded && ARow > 0 && ACol == 0) {
	if ((int) StringGrid1->Objects[IDATABYTE][ARow] & PCM24MASK) {
		StringGrid1->Canvas->Font->Color = clWhite;
#ifdef UNICODE
		StringGrid1->Canvas->Brush->Color = clMedGray;
#else
		StringGrid1->Canvas->Brush->Color = (TColor) 0xA4A0A0;
#endif
		}
	}

if (Decoded && ARow > 0 && ACol > 0 && ((int) StringGrid1->Objects[IDATABYTE][ARow] != 0x00)) {
	// Hightlight the data field if the row contains an index entry
	if (ACol == IDATABYTE) {
		if ((ACol == IDATABYTE) && ((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x00FF) == 0x00) {
			StringGrid1->Canvas->Font->Style = StringGrid1->Canvas->Font->Style << fsStrikeOut;
			}
		if ((int) StringGrid1->Objects[IDATABYTE][ARow] & PCM24MASK) {	//a 24 bit PCM clip
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x00) StringGrid1->Canvas->Font->Color = clFuchsia; // Engine Sound Clip
			// Next line, uses clSkyBlue for CB10.1 and clOlive for BCB4.0
#if __BORLANDC__ >= 0x720		// CB10.1 Only
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x10) StringGrid1->Canvas->Font->Color = clSkyBlue; // Accent Sound Clip
#else
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x10) StringGrid1->Canvas->Font->Color = (TColor) 0xF0CAA6; 	// Accent Sound Clip
#endif
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x20) StringGrid1->Canvas->Font->Color = clRed; 	// Horn/Whistle Sound Clip
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x30) StringGrid1->Canvas->Font->Color = clGreen; 	// Bell Sound Clip
			} else {    //a 4 bit ADPCM clip
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x00) StringGrid1->Canvas->Font->Color = clBlack; 	// Engine Sound Clip
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x10) StringGrid1->Canvas->Font->Color = clAqua; 	// Accent Sound Clip
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x20) StringGrid1->Canvas->Font->Color = clMaroon; 	// Horn/Whistle Sound Clip
			if (((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x30) == 0x30) StringGrid1->Canvas->Font->Color = clLime;	// Bell Sound Clip
			}
		}
#ifdef NYBBLEADDRESSES
#ifdef ODDNYBBLE
	// Highlighting for ADPCM files
	// Highlight the address fields in Green if address was an odd nybble for a 4 bit ADPCM file
	if (!((int) StringGrid1->Objects[IDATABYTE][ARow] & PCM24MASK)) {
		if ((ACol == ISTARTADDR) && ((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x0800)) {
			StringGrid1->Canvas->Font->Color = clGreen;
			}
		if ((ACol == IENDREPADDR) && ((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x0400)) {
			StringGrid1->Canvas->Font->Color = clGreen;
			}
		if ((ACol == ISTARTREPADDR) && ((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x0200)) {
			StringGrid1->Canvas->Font->Color = clGreen;
			}
		if ((ACol == IENDADDR) && ((int) StringGrid1->Objects[IDATABYTE][ARow] & 0x0100)) {
			StringGrid1->Canvas->Font->Color = clGreen;
			}
		}
#else
	// Highlighting for ADPCM files
	// Highlight the address fields in Green if address was an even nybble for a 4 bit ADPCM file
	if (!((int) StringGrid1->Objects[IDATABYTE][ARow] & PCM24MASK)) {
		if ((ACol == ISTARTADDR) && (!(int) StringGrid1->Objects[IDATABYTE][ARow] & 0x0800)) {
			StringGrid1->Canvas->Font->Color = clGreen;
			}
		if ((ACol == IENDREPADDR) && (!(int) StringGrid1->Objects[IDATABYTE][ARow] & 0x0400)) {
			StringGrid1->Canvas->Font->Color = clGreen;
			}
		if ((ACol == ISTARTREPADDR) && (!(int) StringGrid1->Objects[IDATABYTE][ARow] & 0x0200)) {
			StringGrid1->Canvas->Font->Color = clGreen;
			}
		if ((ACol == IENDADDR) && (!(int) StringGrid1->Objects[IDATABYTE][ARow] & 0x0100)) {
			StringGrid1->Canvas->Font->Color = clGreen;
			}
		}
#endif
#endif
	// Highlighting for PCM24 files
	// Highlight the address fields in Red or Green if address is not modulo 3 for a 24 bit signed PCM file
	if (((int) StringGrid1->Objects[IDATABYTE][ARow] & PCM24MASK)) {
		// Use _tcstol for both BCB4 and CB10          see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
		const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][ARow].c_str(), NULL, 16);
		const int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][ARow].c_str(), NULL, 16);
		const int StartRepAddr = _tcstol(StringGrid1->Cells[ISTARTREPADDR][ARow].c_str(), NULL, 16);
		const int EndRepAddr = _tcstol(StringGrid1->Cells[IENDREPADDR][ARow].c_str(), NULL, 16);
		int EndRepDiff = (EndRepAddr - StartAddr) % 3;
		int StartRepDiff = (StartRepAddr - StartAddr) % 3;
		int EndDiff = (EndAddr- StartAddr) % 3;
		//
		if ((ACol == IENDREPADDR) && (EndRepDiff == 1)) { // Use Red if modulo 3 == 1
			StringGrid1->Canvas->Font->Color = clRed;
			} else if ((ACol == IENDREPADDR) && (EndRepDiff == 2)) { // Use Green if modulo 3 == 2
				StringGrid1->Canvas->Font->Color = clGreen;
				}
		if ((ACol == ISTARTREPADDR) && (StartRepDiff == 1)) {
			StringGrid1->Canvas->Font->Color = clRed;
			} else if ((ACol == ISTARTREPADDR) && (StartRepDiff == 2)) {
				StringGrid1->Canvas->Font->Color = clGreen;
				}
		if ((ACol == IENDADDR) && (EndDiff == 1)) {
			StringGrid1->Canvas->Font->Color = clRed;
			} else if ((ACol == IENDADDR) && (EndDiff == 2)) {
				StringGrid1->Canvas->Font->Color = clGreen;
				}
		if ((ACol == ILENGTH) && ((StringGrid1->Cells[ILENGTH][ARow] % 3) == 1)) {  // use Red if PCM clip length is not mod 3
			StringGrid1->Canvas->Font->Color = clRed;
			} else if ((ACol == ILENGTH) && ((StringGrid1->Cells[ILENGTH][ARow] % 3) == 2)) {  // use Green if PCM clip length is not mod 3
				StringGrid1->Canvas->Font->Color = clGreen;
				}
		}
	}
// fill the rectangle
StringGrid1->Canvas->FillRect(Rect);

// Center text
String StringToAlign = StringGrid1->Cells[ACol][ARow];
const int StringStartH = (Rect.Left+Rect.Right)/2 - StringGrid1->Canvas->TextWidth(StringToAlign)/2;
const int StringStartV = (Rect.Top+Rect.Bottom)/2 - StringGrid1->Canvas->TextHeight(StringToAlign)/2;
StringGrid1->Canvas->TextOut(StringStartH, StringStartV, StringToAlign);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1SelectCell(TObject *Sender, int ACol, int ARow,
		  bool &CanSelect)
{
if (ACol != 0 && ARow != 0) CanSelect = true;
}
//---------------------------------------------------------------------------
#ifdef CLICKING
// predValue was declared short
void __fastcall TADPCM_Main::PlayClip(int StartAddr, int StartRepAddr, int EndRepAddr, int EndAddr, int DisplayAddr, short index, int predValue)
#else
void __fastcall TADPCM_Main::PlayClip(int StartAddr, int StartRepAddr, int EndRepAddr, int EndAddr, int DisplayAddr, short index, short predValue)
#endif
{
MyShowMessage("Play Clip: #" +  String(StringGrid1->Row), false);
//
if (EndAddr != -1 && StartAddr >= EndAddr) {
	MyShowMessage("Play Clip: StartAddr >= EndAddr", false);
	return;
	}

if (FormatGroup->ItemIndex == PCM24) {
	if (StartAddr != 0 && StartRepAddr != 0 && (StartAddr % 3) != (StartRepAddr % 3)) {
		MyShowMessage("Clip may not play correctly - Start Address % 3 != Start Repeating Address % 3", true);
		}
	}
HANDLE hFile;/* file handle */
WAVEFORMATEX wfx; /* look this up in your documentation */
/*
 * initialise the module variables
 */
waveBlocks = allocateBlocks(BLOCK_SIZE, BLOCK_COUNT);
waveFreeBlockCount = BLOCK_COUNT;
waveCurrentBlock= 0;
/*
 * try and open the file
 */
if((hFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Play Clip: Unable to open file: " + MRUComboBox->Text, true);
	return;
	}

PlayClipButton->Enabled = false;
EditIndexButton->Enabled = false;
EditClipButton->Enabled = false;
EditSoftkeysButton->Enabled = false;
EditParamsButton->Enabled = false;
ExportOneCliptomp3->Enabled = false;

PlayRepeatingSegmentButton->Enabled = false;
LongPlayButton->Enabled = false;
PlayEntireFileButton->Enabled = false;
PauseButton->Enabled = true;
RestartButton->Enabled = true;
StopButton1->Enabled = true;
StopButton2->Enabled = true;
#ifdef NOTUSEFUL
//
// Display the first 24 bits of the PCM sound - Control Word
//
if (ADPCM_Main->FormatGroup->ItemIndex == PCM24) {	// for PCM24 files, display the first 24 bits control word
	byte cBuffer3[3];								// 24 bit big endian signed PCM Sample
	DWORD readBytes;
	SetFilePointer(hFile, StartAddr, NULL, FILE_BEGIN);
	ReadFile(hFile, cBuffer3, 3, &readBytes, NULL);
	if (readBytes == 3) {
		String CB = HEXFIELDPREFIX + IntToHex((int)cBuffer3[0], 2) + IntToHex((int)cBuffer3[1], 2) + IntToHex((int)cBuffer3[2],2 );
		ADPCM_Main->MyShowMessage("Play Segment: PCM Control Word : " + CB, false);
		}

	}
 #endif

InitializeCriticalSection(&waveCriticalSection);
// locate end of file if needed
if (EndAddr == -1) {
	EndAddr = MyGetFileSizeBytes(hFile) - 1;	// offset of last byte in file (0 based)
	}
/*
 * set up the WAVEFORMATEX structure.
 */
switch(FormatGroup->ItemIndex) {	// these are the same - we send only 16 bit signed PCM to the PC's waveOut routine
	default:
	case -1:
		MyShowMessage("Automatic Settings unable to determine the Clip's Sound Format", true);
		// flow continues to set default
	case ADPCM:	// 4 bit ADPCM converted to 16 bit PCM
	case PCM24:	// 24 bit big endian signed PCM converted to 16 bit PCM
	case PCM16:	// 16 bit signed PCM no conversion needed
		wfx.nSamplesPerSec = SR[SampleGroup->ItemIndex].Value;
		wfx.wBitsPerSample = 16; /* sample size */
		wfx.nChannels= 1; /* channels*/
		wfx.cbSize = 0; /* size of _extra_ info */
		wfx.wFormatTag = WAVE_FORMAT_PCM;
		wfx.nBlockAlign =  (wfx.wBitsPerSample * wfx.nChannels) >> 3; 	// https://www.icculus.org/SDL_sound/downloads/external_documentation/wavecomp.htm
		wfx.nAvgBytesPerSec =   wfx.nBlockAlign * wfx.nSamplesPerSec;	//  see above ref
	}
/*
 * try to open the default wave device. WAVE_MAPPER is
 * a constant defined in mmsystem.h, it always points to the
 * default wave device on the system (some people have 2 or
 * more sound cards).
 */
if((waveOutOpen(&hWaveOut, WAVE_MAPPER, &wfx, (ULONG) waveOutProc, (ULONG) &waveFreeBlockCount , CALLBACK_FUNCTION)) != MMSYSERR_NOERROR) {
	MyShowMessage("Play Clip: Unable to open wave mapper device", true);
	PlayEntireFileButton->Enabled = true;
	PauseButton->Enabled = false;
	RestartButton->Enabled = false;
	StopButton1->Enabled = false;
	StopButton2->Enabled = false;
	StringGrid1Click(NULL);	// set button state
	DeleteCriticalSection(&waveCriticalSection);
	hWaveOut = NULL;
	CloseHandle(hFile);
	return;
	} else {
	//MyShowMessage("Play Clip : Wave mapper device opened", false);
	AudioMin = 32767;
	AudioMax = -32768;
	AudioMin24 = 8388607;
	AudioMax24 = -8388608;
	}
/*
 * playback loop
 */
ImaState PCstate;
BufIndex = BytesSent = BuffersSent = 0;
/*
Notes:
Addresses for 4 bit ADPCM point to exactly the byte of the pair of samples
	The HO nybble has the first sample, the LO nybble has the second sample
Addresses for 24 bit signed PCM point to the first byte of the three bytes for the sample
*/
if ((StartAddr == 0) & (StartRepAddr != 0)) {
  // play the repeating segment
  PCstate.step = index;
  PCstate.predValue = predValue;
  PlaySegment(hWaveOut, hFile, StartRepAddr, EndRepAddr, 0, PCstate);
  }
  else if (StartRepAddr == 0) {
	// a regular clip
	// Play the clip in 1 part
	PCstate.step = 0;
	PCstate.predValue = 0;
	PlaySegment(hWaveOut, hFile, StartAddr, EndAddr, DisplayAddr, PCstate);
	// if its the whistle/horn clip 152, play clip 153
	if (StringGrid1->Row == 152) {
		// get start and end address of ending clip
		// Use _tcstol for both BCB4 and CB10         see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
		int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][StringGrid1->Row + 1].c_str(), NULL, 16);
		int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][StringGrid1->Row + 1].c_str(), NULL, 16);
		PCstate.step = 0;
		PCstate.predValue = 0;
		PlaySegment(hWaveOut, hFile, StartAddr, EndAddr, StartRepAddr, PCstate);
		MyShowMessage("Play Clip : Whistle Ending Clip #" + String(StringGrid1->Row + 1), false);
		}
	} else {
	// a repeating clip
	if ((StartAddr == StartRepAddr) && (EndAddr == EndRepAddr)) {
		// these are equal when the entire clip repeats
		for (int repcount = 0 ; repcount < RepUpDown->Position ;  repcount ++) {
			MyShowMessage("Play Clip : Repeating the Entire Clip", false);
			PCstate.step = 0;
			PCstate.predValue = 0;
			PlaySegment(hWaveOut, hFile, StartAddr, EndAddr, 0, PCstate);
			}
		} else {
		// clip has a repeating segment
		if (RepUpDown->Position == 0) {
			// no repeat, play entire clip once
			MyShowMessage("Play Clip : Entire Clip Once", false);
			PCstate.step = 0;
			PCstate.predValue = 0;
			PlaySegment(hWaveOut, hFile, StartAddr, EndAddr, 0, PCstate);
			} else {
			// 1...n repeats
			// play in sections
			// play beginning piece
			MyShowMessage("Play Clip : Introduction",false);
			PCstate.step = 0;
			PCstate.predValue = 0;
			PlaySegment(hWaveOut, hFile, StartAddr, EndRepAddr, 0, PCstate);
			// play repeating piece
			if (RepUpDown->Position == 1) { // once then done
				MyShowMessage("Play Clip : Conclusion", false);
				PCstate.step = index;
				PCstate.predValue = predValue;
				PlaySegment(hWaveOut, hFile, StartRepAddr, EndAddr, 0, PCstate);
				} else {
				for (int repcount = 1 ; repcount < RepUpDown->Position ; repcount ++) {
					MyShowMessage("Play Clip : Repeating the Repeating Segment", false);
					PCstate.step = index;
					PCstate.predValue = predValue;
					//SM = HEXFIELDPREFIX + IntToHex((int)(state.step << 16) + state.predValue, 6);
					//DebugBuffer->Lines->Append("SM2 " + SM);   // state machine debug print
					PlaySegment(hWaveOut, hFile, StartRepAddr, EndRepAddr, 0, PCstate);
					//SM = HEXFIELDPREFIX + IntToHex((int)(state.step << 16) + state.predValue, 6);
					//DebugBuffer->Lines->Append("SM3 " + SM);   // state machine debug print
					}
				// play ending piece
				MyShowMessage("Play Clip : Conclusion", false);
				PCstate.step = index;
				PCstate.predValue = predValue;
				PlaySegment(hWaveOut, hFile, StartRepAddr, EndAddr, 0, PCstate);
				}

			// if it's a whistle and we are repeating (ie, user pressed [Long Play], play one of the three ending clips
			// The program does not use clip 153 here but the real engine might do that.
			if (StringGrid1->Row == 152) {
				int clip = random(3);  // random number 0-2
				clip++;
				// get start and end address of random ending clip
				// Use _tcstol for both BCB4 and CB10         see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
				int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][StringGrid1->Row + 1 + clip].c_str(), NULL, 16);
				int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][StringGrid1->Row + 1 + clip].c_str(), NULL, 16);
				PCstate.step = 0;
				PCstate.predValue = 0;
				PlaySegment(hWaveOut, hFile, StartAddr, EndAddr, StartRepAddr, PCstate);
				MyShowMessage("Play Clip : Random whistle Ending Clip #" + String(StringGrid1->Row + 1 + clip), false);
				}
			}
		}
	}
#ifdef OLDWAY
// if it's a whistle, play one of the four ending clips
if ((StartAddr != 0) && (StringGrid1->Row == 152) && (Record_16bit_little_endian_PCM == false || ExportOneClipFlag == true)) {
	int clip = random(4);  // random number 0-3
	clip++;
	// get start and end address of random ending clip
	// Use _tcstol for both BCB4 and CB10         see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
	int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][StringGrid1->Row + clip].c_str(), NULL, 16);
	int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][StringGrid1->Row + clip].c_str(), NULL, 16);
	PCstate.step = 0;
	PCstate.predValue = 0;
	PlaySegment(hWaveOut, hFile, StartAddr, EndAddr, StartRepAddr, PCstate);
	MyShowMessage("Play Clip : Random whistle Ending Clip #" + String(StringGrid1->Row + clip), false);
	}
#endif

// Send Last partial Buffer, if any
#ifdef OLDWAY
if (BufIndex != 0) {
	int sizeb2 = sizeof(sBuffer2048); 	// 4096
	memset(sBuffer2048 + BufIndex, '\0', sizeb2 - (BufIndex * 2));
	writeAudio(hWaveOut, (LPSTR) sBuffer2048, sizeof(sBuffer2048));	// 16 bit PCM
	}
#else
if (BufIndex != 0) {
	int sizeb2 = sizeof(sBuffer2048); 	// 4096
	memset(sBuffer2048 + BufIndex, '\0', sizeb2 - (BufIndex * 2));
	writeAudio(hWaveOut, (LPSTR) sBuffer2048, BufIndex * 2);	// 16 bit PCM
	bool RecordFlagSaved = RecordFlag;
	RecordFlag = false; 	// don't record the zero fill bytes
	writeAudio(hWaveOut, (LPSTR) (sBuffer2048 + BufIndex), sizeb2 - (BufIndex * 2));	// 16 bit PCM
	RecordFlag = RecordFlagSaved;
	}
#endif

//MyShowMessage("Play Clip : Last buffer sent to sound card - now playing", false);

const bool RecordFlagSaved = RecordFlag;
RecordFlag = false;     // don't record the quiet buffer of zeros
memset(sBuffer2048, '\0', sizeof(sBuffer2048));	// buffer of quiet needed for very short < 0.10 second clips
												// and for the end fraction of second of every clip
writeAudio(hWaveOut, (LPSTR) sBuffer2048, sizeof(sBuffer2048));	// 16 bit PCM
RecordFlag = RecordFlagSaved;

TCHAR str[25], str2[25];
float ClipSeconds = BytesSent/((float)wfx.nSamplesPerSec * 2.0);
_sntprintf(str, sizeof str, TEXT("%6.2f"), ClipSeconds);
MyShowMessage("Play Clip : Audio Bytes Sent (2 samples/byte) = " + String(BytesSent) + " - Buffers = " + String(BuffersSent), false);
MyShowMessage("Play Clip : Length = " + String(str) + " seconds", false);
_sntprintf(str, sizeof str, TEXT("%6.2f"), (AudioMax * 100)/32767.0);
_sntprintf(str2, sizeof str, TEXT("%6.2f"), (AudioMin * 100)/32767.0);
MyShowMessage("Play Clip : 16 bit Dynamic Range = " + String(AudioMax) + " to " + String(AudioMin) + " (" + String(str) + "% to " + String(str2) + "%)", false);
if (iDebug > 0) {
	_sntprintf(str, sizeof str, TEXT("%6.2f"), (AudioMax24 * 100)/(8388607.0));
	_sntprintf(str2, sizeof str, TEXT("%6.2f"), (AudioMin24 * 100)/(8388607.0));
	MyShowMessage("Play Clip : 24 bit Dynamic Range = " + String(AudioMax24) + " to " + String(AudioMin24) + " (" + String(str) + "% to " + String(str2) + "%)", false);
	}
/*
 * wait for all blocks to complete
 */
while(waveFreeBlockCount < BLOCK_COUNT) {
	Sleep(1);
	Application->ProcessMessages();
	}

/*
 * unprepare any blocks that are still prepared
 */
for(int i = 0; i < waveFreeBlockCount; i++)
	if(waveBlocks[i].dwFlags & WHDR_PREPARED)
		if (waveOutUnprepareHeader(hWaveOut, &waveBlocks[i], sizeof(WAVEHDR)) != MMSYSERR_NOERROR) {
			MyShowMessage("Play Clip : waveOutUnprepareHeader Failure",false);
			}
DeleteCriticalSection(&waveCriticalSection);

freeBlocks(waveBlocks);
if (waveOutClose(hWaveOut) != MMSYSERR_NOERROR) {
	MyShowMessage("Play Clip : waveOutClose Failure", false);
	}

hWaveOut = NULL;
CloseHandle(hFile);

//MyShowMessage("Play Clip : Wave mapper device closed",false);
PlayEntireFileButton->Enabled = true;
PauseButton->Enabled = false;
RestartButton->Enabled = false;
StopButton1->Enabled = false;
StopButton2->Enabled = false;
PlayClipButton->Enabled = true;
EditIndexButton->Enabled = true;
EditClipButton->Enabled = true;
EditSoftkeysButton->Enabled = true;
EditParamsButton->Enabled = true;
ExportOneCliptomp3->Enabled = true;

if (StringGrid1->Cells[IENDREPADDR][StringGrid1->Row] == "") {
	PlayRepeatingSegmentButton->Enabled = false;
	LongPlayButton->Enabled = false;
	} else {
	PlayRepeatingSegmentButton->Enabled = true;
	LongPlayButton->Enabled = true;
	}
return;
}
//---------------------------------------------------------------------------

void __fastcall PlaySegment(HWAVEOUT hWaveOut, HANDLE hSoundFile, int StartAddr, int EndAddr, int DisplayAddr, ImaState & PSstate)
{
//
// This routine plays a simple clip from a Starting Address to an Ending Address
// State machine initialization is used if its non-zero. It should be zero for
// a plain sound clip. It might be non-zero if the clip is a repeating segment.
// In any case, the state machine init is only used for ADPCM clips.
//
if (EndAddr != -1 && StartAddr > EndAddr) {
    // don't try to play if the addresses are wrong
	//ADPCM_Main->MyShowMessage("Play Segment: StartAddr > EndAddr", true);
	return;
	}

unsigned char cBuffer3[3];	// holds 2 4 bit ADPCM Samples or 1 24 bit big endian signed PCM Sample

// set file pointer to first byte of the seqment
const long  fp = StartAddr;
SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
if (ADPCM_Main->FormatGroup->ItemIndex == ADPCM) {
	// Display IMA State Machine
	String SM = HEXFIELDPREFIX + IntToHex((int)((PSstate.step & 0x00FF) << 16) | (PSstate.predValue & 0x00FFFF), 6);
	String MAS = HEXFIELDPREFIX + IntToHex((int)StartAddr, 6);
	ADPCM_Main->MyShowMessage("Play Segment: ADPCM Play Clip at Enter : SM = " + SM + " (" + MAS + ")", false);
	}
if (ADPCM_Main->FormatGroup->ItemIndex == PCM24 || ADPCM_Main->FormatGroup->ItemIndex == PCM16) {
	String MAS = HEXFIELDPREFIX + IntToHex((int)StartAddr, 6);
	ADPCM_Main->MyShowMessage("Play Segment: PCM Play Clip at Enter (" + MAS + ")", false);
	}
// set length of seqment
const int len = EndAddr - StartAddr + 1;

Stopem = false;
DWORD readBytes;
int TotalBytes = len;
int iBCount = 0;
bool Done = false;
//String SM, MAS;
//
while(1) {
	int16_t Value1a, Value1b, Value16;
	int32_t Value24, Value32;
	if (Done)	break;	// done
	//---
	switch(ADPCM_Main->FormatGroup->ItemIndex) {
        default:
		case ADPCM:		// 4 bit ADPCM converted to 16 bit signed PCM
			// Decompress IMA ADPCM to 16 bit PCM, adjust volume and write to Audio

			// Next block of code moved here from a few lines down  v1.5.7
			// Display the State Machine initialization value
			if ((StartAddr + iBCount) == DisplayAddr) {
				String SM = HEXFIELDPREFIX + IntToHex((int)((PSstate.step & 0x00FF) << 16) | (PSstate.predValue & 0x00FFFF), 6);
				String MAS = HEXFIELDPREFIX + IntToHex((int)StartAddr + iBCount, 6);
				ADPCM_Main->MyShowMessage("Play Segment: ADPCM SM Init = " + SM + " (" + MAS + ")", false);
				}
							
			if(!ReadFile(hSoundFile, cBuffer3, 1, &readBytes, NULL)) break;
			if(readBytes !=  1) {
				//MyShowMessage("Read zero bytes");
				Done = true;
				break; 	// hit end of file
				}
			iBCount++;	// count bytes read from sound file
			TotalBytes -= readBytes;	// subtract bytes read if it goes negative, we are done
			Value1a = (ImaAdpcmDecode((cBuffer3[0] & 0xF0) >> 4, PSstate));
        	//Value1a /= 2;	// DEBUG volume level reduction
			writeSample(hWaveOut, Value1a);
			Value1b = (ImaAdpcmDecode((cBuffer3[0] & 0x0F), PSstate));
        	//Value1b /= 2;	// DEBUG volume level reduction
			writeSample(hWaveOut, Value1b);

			// code here to display SM Init value moved a few lines up v1.5.7 
			
			if (ADPCM_Main->iDebug > 1) {
				// Debug - show each 16 bit sound value sent to the sound card
				ADPCM_Main->MyShowMessage(String(iBCount) + " " + HEXFIELDPREFIX + IntToHex(Value1a, 4) + ")", false);
				ADPCM_Main->MyShowMessage(String(iBCount) + " " + HEXFIELDPREFIX + IntToHex(Value1b, 4) + ")", false);
				}

			// End Debug
			if (TotalBytes <= 0) {	// all done after this buffer is processed
				Done = true;
				break;
				}

			break;
		case PCM24:		// 24 bit signed PCM converted to 16 bit signed PCM
			if(!ReadFile(hSoundFile, cBuffer3, 3, &readBytes, NULL)) break;	// read 24 bits
			if(readBytes != 3) {
				Done = true;
				break;	// out of for loop - need three bytes to process
				}
			TotalBytes -= readBytes;	// subtract bytes read if it goes negative, we are done
			// Big Endian
			Value16 = ((cBuffer3[0] << 8) | cBuffer3[1]);	// 24 bit big endian to 16 bit signed PCM
			// use high order bit of cBuffer3[2] to round the value
			if (cBuffer3[2] & 0x80) Value16++;
						
			if (ADPCM_Main->iDebug > 2) {
				// Debug - show each 24 bit sound value sent to the engine
				// https://stackoverflow.com/questions/42534749/signed-extension-from-24-bit-to-32-bit-in-c
				int32_t Value32 = (int8_t) cBuffer3[0]; 			  // C guarantees the sign extension
				Value32 <<= 16; 									  // shift the byte at its definitive place
				Value32 |= ((int32_t) (uint8_t) cBuffer3[1]) << 8;    // place the second byte
				Value32 |= ((int32_t) (uint8_t) cBuffer3[2]);         // and the least significant byte
				if (Value32 > ADPCM_Main->AudioMax24) {
                	ADPCM_Main->AudioMax24 = Value32;	// retain min and max audio value for debugging
                    }
				if (Value32 < ADPCM_Main->AudioMin24) {
                	ADPCM_Main->AudioMin24 = Value32;
                    }
				ADPCM_Main->MyShowMessage(String(iBCount) + " " + HEXFIELDPREFIX + IntToHex(Value32 & (int32_t) 0x00FFFFFF, 6) + " (" + String(Value32) + ")", false);
				}				

			if (ADPCM_Main->iDebug > 1) {
				// Debug - show each 16 bit sound value sent to the sound card
				// https://stackoverflow.com/questions/42534749/signed-extension-from-24-bit-to-32-bit-in-c
				int16_t Value16 = (int8_t) cBuffer3[0]; 			  // C guarantees the sign extension
				Value16 <<= 8; 										  // shift the byte at its definitive place
				Value16 |= ((int32_t) (uint8_t) cBuffer3[1]);         // and the least significant byte			
				ADPCM_Main->MyShowMessage(String(iBCount) + " " + HEXFIELDPREFIX + IntToHex(Value16 & (int32_t) 0x0000FFFF, 4) + " (" + String(Value32) + ")", false);
				}

			// the rest of cBuffer3[2] is discarded as we don't need the low bits
			writeSample(hWaveOut, Value16);
			iBCount += 3;	// count bytes read from sound file
			if (TotalBytes <= 0) {	// all done after this buffer is processed
				Done = true;
				break;
				}
			break;
		case PCM16: // 16 bit signed PCM no conversion needed
			if(!ReadFile(hSoundFile, cBuffer3, 2, &readBytes, NULL)) break;	// read 16 bits
			if(readBytes != 2) {
				Done = true;
				break;	// out of for loop - need two bytes to process
				}
			TotalBytes -= readBytes;	// subtract bytes read if it goes negative, we are done
			Value16 = ((cBuffer3[1] << 8) | cBuffer3[0]);	// assemble 16 bit sample - little endian
			writeSample(hWaveOut, Value16);
			iBCount += 2;	// count bytes read from sound file
			if (TotalBytes <= 0) {	// all done after this buffer is processed
				Done = true;
				break;
				}
			break;
        }
	if(Done) break;
	//---
	Application->ProcessMessages();
	if (Stopem) break;	// cancel
	}
if (ADPCM_Main->FormatGroup->ItemIndex == ADPCM) {
	// Display IMA State Machine
	String SM = HEXFIELDPREFIX + IntToHex((int)((PSstate.step & 0x00FF) << 16) | (PSstate.predValue & 0x00FFFF), 6);
	String MAS = HEXFIELDPREFIX + IntToHex((int)StartAddr + iBCount, 6);
	ADPCM_Main->MyShowMessage("Play Segment: ADPCM Play Clip at Exit : SM = " + SM + " (" + MAS + ")", false);
	}
if (ADPCM_Main->FormatGroup->ItemIndex == PCM24 || ADPCM_Main->FormatGroup->ItemIndex == PCM16) {
	String MAS = HEXFIELDPREFIX + IntToHex((int)StartAddr + iBCount, 6);
	ADPCM_Main->MyShowMessage("Play Segment: PCM Play Clip at Exit (" + MAS + ")", false);
	}
}
//---------------------------------------------------------------------------

void writeSample(HWAVEOUT hWaveOut, short Value)
{
//
//Value *= 14;  // amplify - debugging only
//
sBuffer2048[BufIndex] = Value;
//
if (Value > ADPCM_Main->AudioMax) {
	ADPCM_Main->AudioMax = Value;	// retain min and max audio value for debugging
    }
if (Value < ADPCM_Main->AudioMin) {
	ADPCM_Main->AudioMin = Value;
    }
//
BufIndex++;
ADPCM_Main->BytesSent += 2;
if (BufIndex == 2048) {
	// write it
	writeAudio(hWaveOut, (LPSTR) sBuffer2048, sizeof(sBuffer2048));	// 16 bit PCM
	BufIndex = 0;	// reset
	return;
	}
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::LongPlayButtonClick(TObject *Sender)
{
// play clip repeating any repeating section based on value of RepUpDown->Position
Overflow->Visible = false;
Overflow->Checked = false;
// pick out the IMA ADPCM state machine's index and previous value
// Use _tcstol for both BCB4 and CB10         see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const short index = _tcstol(StringGrid1->Cells[ISMINIT][StringGrid1->Row].c_str(), NULL, 16) >> 16;
const short predValue = _tcstol(StringGrid1->Cells[ISMINIT][StringGrid1->Row].c_str(), NULL, 16) & 0x0000FFFF;

if (index < -1 || index > 88) {
	MyShowMessage("IMA ADPCM index out of range: " + String(index), false);
	return;
	}

if (RepUpDown->Position == 0) {
	MyShowMessage("For [Long Play], Repeat Count must be > 0", true);
	return;
	}

TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

// Use _tcstol for both BCB4 and CB10         see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][StringGrid1->Row].c_str(), NULL, 16);
const int EndRepAddr = _tcstol(StringGrid1->Cells[IENDREPADDR][StringGrid1->Row].c_str(), NULL, 16);
int StartRepAddr = _tcstol(StringGrid1->Cells[ISTARTREPADDR][StringGrid1->Row].c_str(), NULL, 16);
const int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][StringGrid1->Row].c_str(), NULL, 16);
if (FormatGroup->ItemIndex == PCM24) {   // 24 pit PCM
	//StartRepAddr = StartRepAddr - abs((StartRepAddr % 3) - (StartAddr % 3));
	if ((StartRepAddr % 3) != (StartAddr % 3)) StartRepAddr++;
	if ((StartRepAddr % 3) != (StartAddr % 3)) StartRepAddr++;
	}

// (Start Address, Start Repeat Address, End Repeat Address, End Address, state.step, state.predValue)
PlayClip(StartAddr, StartRepAddr, EndRepAddr, EndAddr, 0, index, predValue);
Screen->Cursor = OldCursor;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::Clear1Click(TObject *Sender)
{
DebugBuffer->Clear();
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::Bell_IndexClick(TObject *Sender)
{
StringGrid1->Row = 157; // 0xAC0
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::Toot_IndexClick(TObject *Sender)
{
StringGrid1->Row = 152; // 0xA70
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::PFA_IndexClick(TObject *Sender)
{
StringGrid1->Row = 80; // 0x5F0
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::Clickity_Clack_IndexClick(TObject *Sender)
{
StringGrid1->Row = 46; // 0x3D0
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::EngineSounds_IndexClick(TObject *Sender)
{
StringGrid1->Row = 177; // 0xC00
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::SFS_IndexClick(TObject *Sender)
{
StringGrid1->Row = 43; // 0x3A0
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::SXS_IndexClick(TObject *Sender)
{
StringGrid1->Row = 255; // 0x10E0
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ClipInfoClick(TObject *Sender)
{
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
// Display Clip info depending on clip type
const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][StringGrid1->Row].c_str(), NULL, 16);    // Start
if (StartAddr == 0) return;	// NOP
const int StartRepAddr = _tcstol(StringGrid1->Cells[ISTARTREPADDR][StringGrid1->Row].c_str(), NULL, 16);  // Start Rep
const int EndRepAddr = _tcstol(StringGrid1->Cells[IENDREPADDR][StringGrid1->Row].c_str(), NULL, 16);    // End Rep
const int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][StringGrid1->Row].c_str(), NULL, 16);    // End
const int Length =  StringGrid1->Cells[ILENGTH][StringGrid1->Row].ToInt();    // Length base 10
const int databyte =   _tcstol(StringGrid1->Cells[IDATABYTE][StringGrid1->Row].c_str(), NULL, 16);    // data byte
//
// display filename
//
MyShowMessage(MRUComboBox->Text, false);

MyShowMessage("Clip # = " + StringGrid1->Cells[ICLIPNO][StringGrid1->Row], false);
if (FormatGroup->ItemIndex == ADPCM) {   // 4 bit ADPCM
	MyShowMessage("4 bit ADPCM format", false);
	} else {
	MyShowMessage("24 bit big endian PCM format", false);
	}
MyShowMessage("StartAddr = " + String(StartAddr), false);
MyShowMessage("EndAddr = " + String(EndAddr), false);
if (StartRepAddr != 0) {
	MyShowMessage("StartRepAddr = " + String(StartRepAddr), false);
	MyShowMessage("EndRepAddr = " + String(EndRepAddr), false);
	}
MyShowMessage("Length = " + String(Length), false);
MyShowMessage("Data Byte = " + String(databyte), false);

if (FormatGroup->ItemIndex == ADPCM) {   // 4 bit ADPCM
	MyShowMessage("State Machine Init = " + StringGrid1->Cells[ISMINIT][StringGrid1->Row], false);
	}
if (FormatGroup->ItemIndex == PCM24) {   // 24 pit PCM
	MyShowMessage("StartAddr % 3 = " + String(StartAddr % 3), false);
	MyShowMessage("EndAddr % 3 = " + String(EndAddr % 3), false);
	MyShowMessage("(StartAddr - EndAddr) % 3 = " + String((EndAddr - StartAddr) % 3), false);
	if (StartRepAddr != 0) {
		MyShowMessage("StartRepAddr % 3 = " + String(StartRepAddr % 3), false);
		MyShowMessage("EndRepAddr % 3 = " + String(EndRepAddr % 3), false);
		MyShowMessage("(StartRepAddr - EndRepAddr) % 3 = " + String((EndRepAddr - StartRepAddr) % 3), false);
		}
	MyShowMessage("Unknown Data Word = " + StringGrid1->Cells[ISMINIT][StringGrid1->Row], false);
	}
MyShowMessage(" ", false);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ExportOneCliptomp3Click(TObject *Sender)
{
// Export one Clip
//ExportOneClipFlag  = true;
ExportClip(StringGrid1->Row, false);
//ExportOneClipFlag  = false;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::Export1Cliptomp3Click(TObject *Sender)
{
// Export one Clip
//ExportOneClipFlag  = true;
ExportClip(StringGrid1->Row, false);
//ExportOneClipFlag  = false;
}
//---------------------------------------------------------------------------
void __fastcall TADPCM_Main::ExportClip(int index, bool DFPlayerFormat)
{
// Export one clip
RecordFlag = true;
const SHORT ShiftKeyState = GetKeyState(VK_SHIFT);		// if Shift is pressed, call LongPlayButtonClick()
// set index into StringGrid
StringGrid1->Row = index;
// Use _tcstol for both BCB4 and CB10          see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][StringGrid1->Row].c_str(), NULL, 16);
const int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][StringGrid1->Row].c_str(), NULL, 16);
if (StartAddr == 0x00) {
	RecordFlag = false;
	return;	// empty index slot
	}
// create temporary filename
// create file
String NewExtension =  "_" + FormatFloat("000", index) + "_16bit_little_endian_PCM_" + SR[SampleGroup->ItemIndex].Hz + ".raw";
const String RecordFileName = ChangeFileExt(MRUComboBox->Text, NewExtension);

// delete recording file if present
if (FileExists(RecordFileName))
	DeleteFile(RecordFileName);

// open temporary output file which will be created as 16 bit little endian signed PCM
if((hRawFile = CreateFile(RecordFileName.c_str(), GENERIC_WRITE, NULL, NULL, CREATE_ALWAYS, 0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Export Clip: Unable to open file: " + RecordFileName, true);
	RecordFlag = false;
	return;
	} else {
	MyShowMessage("Export Clip: Exporting to file: " + RecordFileName, false);
	}
// play it and write the output to a file
if (ShiftKeyState < 0) {
	// shifted, call LongPlayButtonClick
	LongPlayButtonClick(NULL);
	} else {
	// unshifted, call PlayClip
	PlayClip(StartAddr, 0, 0, EndAddr, 0, 0, 0);
	}

// close temporary output file
CloseHandle(hRawFile);

// Build Command Line
String OutputFileName;
if (!DFPlayerFormat) {
	// Normal Filename
	NewExtension = "_" + FormatFloat("000", index) + ".mp3";
	OutputFileName = ChangeFileExt(MRUComboBox->Text, NewExtension);
	} else {
	// DF Player Filename - place on SD card in folder /mp3
    // MUST be 4 digits with leading zeros
	OutputFileName = FormatFloat("0000", index) + ".mp3";
	}

const String CmdLine = "\"" + String(cwd) + "\\lame.exe\"" ;
// Adding -V 0 creates a larger output file with not much impovement in sound
// output for audio at 5512 Hz is always upsampled to 8000 Hz,
// output at & over 8000 Hz is sampled at its actual rate
// use --resample 22.05  to get higher sample rate for all clips (makes much larger files)
String ParamLine = " --silent -r -s " + String(SR[SampleGroup->ItemIndex].Kz) + " --bitwidth 16 --signed -m m " +
	"\"" + RecordFileName + "\" \"" + OutputFileName + "\"";

MyShowMessage("Export Clip: Executing: " + CmdLine + " " + ParamLine, false);

// delete output file if present
if (FileExists(OutputFileName))
	DeleteFile(OutputFileName);

// working directory
const String WD = ExtractFilePath(MRUComboBox->Text);

// run lame.exe
SHELLEXECUTEINFO ShExecInfo = {0};
ShExecInfo.cbSize = sizeof(SHELLEXECUTEINFO);
ShExecInfo.fMask = SEE_MASK_NOCLOSEPROCESS;
ShExecInfo.hwnd = NULL;
ShExecInfo.lpVerb = NULL;
ShExecInfo.lpFile = CmdLine.c_str();
ShExecInfo.lpParameters = ParamLine.c_str();
ShExecInfo.lpDirectory = WD.c_str();
ShExecInfo.nShow = SW_HIDE;
ShExecInfo.hInstApp = NULL;

ShellExecuteEx(&ShExecInfo);

WaitForSingleObject(ShExecInfo.hProcess,INFINITE);

// delete recording file since we are done with it
if (FileExists(RecordFileName))
	DeleteFile(RecordFileName);

RecordFlag = false;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::Mod3ButtonClick(TObject *Sender)
{
// show the table as mod 3 for the PCM clips
for (int i = 1 ; i <= 255  ; i++) {
	const int databyte = _tcstol(StringGrid1->Cells[IDATABYTE][i].c_str(), NULL, 16);    // data byte
	if (!(databyte & PCM24MASK)) continue;	// skip ADPCM clips
	const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][i].c_str(), NULL, 16);    // Start
	if (StartAddr == 0) continue;       // skip empty index locations
	const int StartRepAddr = _tcstol(StringGrid1->Cells[ISTARTREPADDR][i].c_str(), NULL, 16);  // Start Rep
	const int EndRepAddr = _tcstol(StringGrid1->Cells[IENDREPADDR][i].c_str(), NULL, 16);    // End Rep
	const int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][i].c_str(), NULL, 16);    // End

	StringGrid1->Cells[ISTARTADDR][i] = StartAddr % 3;

	if (StartRepAddr != 0) {
		StringGrid1->Cells[ISTARTREPADDR][i] = StartRepAddr % 3;
		}

	if (EndRepAddr != 0) {
		StringGrid1->Cells[IENDREPADDR][i] = EndRepAddr % 3;
		}

	StringGrid1->Cells[IENDADDR][i] = EndAddr % 3;
	}
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::EditClipClick(TObject *Sender)
{
// edit the clip given by StringGrid1->Row - used by Menu Entry and Button
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
String Extension = LowerCase(ExtractFileExt(MRUComboBox->Text));
if (Extension != ".mth") {
	MyShowMessage("Edit Clip: Can only edit '.mth' files", true);
	return;
	}

int clip = StringGrid1->Row;
const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][clip].c_str(), NULL, 16);    // Start
const int databyte = _tcstol(StringGrid1->Cells[IDATABYTE][clip].c_str(), NULL, 16);    // data byte
// There must be a clip there for Edit, otherwise its an Add
if (StartAddr == 0 || StartAddr == EMPTY_FIELD) {
	clip = -clip;
    }
//
if (Clip_Editor == NULL) {
	Clip_Editor = new TClip_Editor(Application, clip, MRUComboBox->Text);
	assert(Clip_Editor != NULL);
	}

Clip_Editor->ShowIt(clip, MRUComboBox->Text, databyte, EndAddrMax, PSGroup->ItemIndex);	// Show Clip Editor Window
}
//---------------------------------------------------------------------------
void __fastcall TADPCM_Main::MyShowMessage(String Msg, bool Popup)
{
// display a message in the message window and, optionaly, show it in a popup
DebugBuffer->Lines->Append(Msg);
if (Popup) ShowMessage(Msg);
}
//---------------------------------------------------------------------------
void __fastcall TADPCM_Main::MyShowLastError(String Msg, bool Popup)
{
// display a message in the message window along with the GetLastError() result and, optionaly, show it in a popup
//
// Retrieve the system error message for the last-error code
LPVOID lpMsgBuf;
const DWORD dw = GetLastError();

FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dw,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsgBuf,
		0, NULL );
// Convert the error message into a String
String MsgBuf ((LPCTSTR)lpMsgBuf);
LocalFree(lpMsgBuf);
// Display the message and error message
DebugBuffer->Lines->Append(Msg + " " + MsgBuf);
if (Popup) ShowMessage(Msg + " " + MsgBuf);
}
//---------------------------------------------------------------------------
void __fastcall TADPCM_Main::EditIndexClick(TObject *Sender)
{
// edit the data byte
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clip = StringGrid1->Row;
const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][clip].c_str(), NULL, 16);    // Start
const int StartRepAddr = _tcstol(StringGrid1->Cells[ISTARTREPADDR][clip].c_str(), NULL, 16);    // Start Repeat
const int EndRepAddr = _tcstol(StringGrid1->Cells[IENDREPADDR][clip].c_str(), NULL, 16);    // End Repeat
const int SMInit = _tcstol(StringGrid1->Cells[ISMINIT][clip].c_str(), NULL, 16);    // State Machine Init
const int databyte = _tcstol(StringGrid1->Cells[IDATABYTE][clip].c_str(), NULL, 16);    // data byte
// There must be a clip there
if (StartAddr == 0 || StartAddr == EMPTY_FIELD) return;	// NOP
// read the sound file index entry
HANDLE hSoundFile;
SFIndex IndexSound;
// try and open the file
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Old Edit Index: Unable to open file: " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	} else {
	MyShowMessage("OldEditIndex: Sound file to be edited: " + MRUComboBox->Text, false);
	}
if (!ParseIndex(IndexSound, hSoundFile, clip)) {
	MyShowMessage("Old Edit Index: Parse Index failed", true);
	CloseHandle(hSoundFile);
	return;
	}

String Answer = IntToHex(StartRepAddr, 6);
if (StartRepAddr == 0) Answer = "7FFFFF";
if(InputQuery("Old Edit Index", "Enter Value for Start Repeat Address: ", Answer)) {
	StringGrid1->Cells[ISTARTREPADDR][clip] = HEXFIELDPREFIX + Answer;
	IndexSound.StartRepAddr = _tcstol(StringGrid1->Cells[ISTARTREPADDR][clip].c_str(), NULL, 16);
	// write the sound file index entry
	if (!WriteIndex(IndexSound, hSoundFile, clip)) {
		MyShowMessage("Old Edit Index: Could not write index to sound file", true);
		}
	}
Answer = IntToHex(EndRepAddr, 6);
if (EndRepAddr == 0) Answer = "7FFFFF";
if(InputQuery("Old Edit Index", "Enter Value for End Repeat Address: ", Answer)) {
	StringGrid1->Cells[IENDREPADDR][clip] = HEXFIELDPREFIX + Answer;
	IndexSound.EndRepAddr = _tcstol(StringGrid1->Cells[IENDREPADDR][clip].c_str(), NULL, 16);
	// write the sound file index entry
	if (!WriteIndex(IndexSound, hSoundFile, clip)) {
		MyShowMessage("Old Edit Index: Could not write index to sound file", true);
		}
	}
Answer = IntToHex(SMInit, 6);
if(InputQuery("Old Edit Index", "Enter Value for State Machine Init: ", Answer)) {
	StringGrid1->Cells[ISMINIT][clip] = HEXFIELDPREFIX + Answer;
	IndexSound.SMInit = _tcstol(StringGrid1->Cells[ISMINIT][clip].c_str(), NULL, 16);
	// write the sound file index entry
	if (!WriteIndex(IndexSound, hSoundFile, clip)) {
		MyShowMessage("Old Edit Index: Could not write index to sound file", true);
		}
	}
Answer = IntToHex(databyte, 2);
if(InputQuery("Old Edit Index", "Enter Value for Data Byte: ", Answer)) {
	StringGrid1->Cells[IDATABYTE][clip] = HEXFIELDPREFIX + Answer;
	IndexSound.databyte = _tcstol(StringGrid1->Cells[IDATABYTE][clip].c_str(), NULL, 16);
	// write the sound file index entry
	if (!WriteIndex(IndexSound, hSoundFile, clip)) {
		MyShowMessage("Old Edit Index: Could not write index to sound file", true);
		}
	}
CloseHandle(hSoundFile);
AnalyzeButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::PSGroupClick(TObject *Sender)
{
if (AutoCheckBox->Checked) AutoCheckBoxClick(Sender);	// Automatic settings if requested
}
//---------------------------------------------------------------------------


void __fastcall TADPCM_Main::MinimizeFileLenClick(TObject *Sender)
{
// set file length to minimum based on EndAddrMax
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
HANDLE hSoundFile;
// try and open the file
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Minimize File Len: Unable to open file: " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	} else {
	MyShowMessage("Minimize File Len: Sound file to be minimized: " + MRUComboBox->Text, false);
	}

const long fp = EndAddrMax + 1;
SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
SetEndOfFile(hSoundFile);
//
getFileDetails(hSoundFile);
CloseHandle(hSoundFile);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::CopyrightClick(TObject *Sender)
{
// enable debugging by making usually invisible buttons visible
Mod3Button->Visible = true;
EditIndex->Visible = true;
EditIndexButton->Visible = true;
EditClipButton->Visible = true;
EditSoftkeysButton->Visible = true;
EditParamsButton->Visible = true;
ExportOneCliptomp3->Visible = true;
iDebug++;
MyShowMessage("Debug = " + String(iDebug), false);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::EditIndexButtonClick(TObject *Sender)
{
// edit the data byte
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clip = StringGrid1->Row;
// you can edit an empty index location, unlimited editing if no sound file size checks is on.
if (Edit_Index == NULL) {
	Edit_Index = new TEdit_Index(Application, MRUComboBox->Text, (NoSoundFileSizeChecks->Checked) ? 0xFFFFFF : EndAddrMax, clip);
	assert(Edit_Index != NULL);
	}

Edit_Index->ShowIt(MRUComboBox->Text, (NoSoundFileSizeChecks->Checked) ? 0xFFFFFF : EndAddrMax, clip);
}
//---------------------------------------------------------------------------

bool __fastcall TADPCM_Main::CheckEEPROMSize(int Movement)
{
if (NoSoundFileSizeChecks->Checked) {
	MyShowMessage("Sound file size checking is disabled", false);
	MyShowMessage("Sound files larger than the size of memory in engine will not function correctly", false);
	return true;	// disregard file size checks
	}
// Plus movement means source clip is shrunk
// Minus movement means source clip is grown
if (Movement >= 0) return true;	// size is shrinking or there is no change
if (abs(Movement) > AvailableSpace) return false;	// not enough space
return true;	// OK, enough space
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::EditSoftKeysClick(TObject *Sender)
{
String Extension = LowerCase(ExtractFileExt(MRUComboBox->Text));
if (Extension != ".mth") {
	MyShowMessage("Edit Soft Keys: Can only edit '.mth' files", true);
	return;
	}
// Soft Key Editor is Auto Created
SoftKey_Editor->ShowIt(MRUComboBox->Text);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::CopyIndexClick(TObject *Sender)
{
//Copy this index entry to another index
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clipFrom = StringGrid1->Row;
const String FromCaption = "Copy Index from #" + String(clipFrom);

// read the sound file index entry
HANDLE hSoundFile;
SFIndex IndexSound;
// try and open the file
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Copy Index: Unable to open file: " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	} else {
	MyShowMessage("Copy Index: Sound file to be edited: " + MRUComboBox->Text, false);
	}
if (!ParseIndex(IndexSound, hSoundFile, clipFrom)) {
	MyShowMessage("Copy Index: Parse Index failed", true);
	CloseHandle(hSoundFile);
	return;
	}

String Answer = 0;
if(InputQuery(FromCaption, "Enter Destination Index Number (1-255): ", Answer)) {
	const int clipTo = Answer.ToInt();
	if (clipTo > 255 || clipTo < 1) {
		MyShowMessage("Clip # cannot be less than 1 or greater than 255", true);
		CloseHandle(hSoundFile);
		return;
		}
	const String CopyMsg = "Copy Index: Copy Index #" + String(clipFrom) + " to #" + String(clipTo);
	const int retv = Application->MessageBox(CopyMsg.c_str(), TEXT("Confirmation"), MB_OKCANCEL | MB_DEFBUTTON2);
	if (retv == IDCANCEL) {
		CloseHandle(hSoundFile);
		return;
		}

	MyShowMessage(CopyMsg, false);
	// write the sound file index entry
	if (!WriteIndex(IndexSound, hSoundFile, clipTo)) {
		MyShowMessage("Copy Index: Could not write index to sound file", true);
		}
	}
CloseHandle(hSoundFile);
AnalyzeButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ClearthisIndexClick(TObject *Sender)
{
// Clear this index entry
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clip = StringGrid1->Row;
// read the sound file index entry
HANDLE hSoundFile;
SFIndex IndexSound;
// try and open the file
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Clear Index: Unable to open file: " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	} else {
	MyShowMessage("Clear Index: Sound file to be edited: " + MRUComboBox->Text, false);
	}
if (!ParseIndex(IndexSound, hSoundFile, clip)) {
	MyShowMessage("Clear Index: Parse Index failed", true);
	CloseHandle(hSoundFile);
	return;
	}

const String ClearMsg = "Clear Index: Clear Index #" + String(clip);
const int retv = Application->MessageBox(ClearMsg.c_str(), TEXT("Confirmation"), MB_OKCANCEL | MB_DEFBUTTON2);
if (retv == IDCANCEL) {
	CloseHandle(hSoundFile);
	return;
	}

IndexSound.Clip = clip;
IndexSound.StartAddr = EMPTY_FIELD;
IndexSound.EndRepAddr = EMPTY_FIELD;
IndexSound.StartRepAddr = EMPTY_FIELD;
IndexSound.SMInit = 0x000000;
IndexSound.EndAddr = EMPTY_FIELD;
IndexSound.databyte = 0x00;
IndexSound.SA_LOBit = 1;
IndexSound.SRA_LOBit = 1;
IndexSound.ERA_LOBit = 1;
IndexSound.EA_LOBit = 1;

// write the sound file index entry
if (!WriteIndex(IndexSound, hSoundFile, clip)) {
	MyShowMessage("Clear Index: Could not write index to sound file", true);
	}
CloseHandle(hSoundFile);

MyShowMessage(ClearMsg, false);

AnalyzeButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::NoSoundFileSizeChecksClick(TObject *Sender)
{
// toggle NoSoundFileSizeChecks
NoSoundFileSizeChecks->Checked = !NoSoundFileSizeChecks->Checked;
if (NoSoundFileSizeChecks->Checked) {
	EngineEEPROMSize->Font->Color = clRed;
	} else {
	EngineEEPROMSize->Font->Color = clGreen;
	}
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::AvailableTimeRemainingClick(TObject *Sender)
{
// Display remaining time left
TCHAR str[50];
if (FormatGroup->ItemIndex == PCM24) {   // 24 pit PCM
	_sntprintf(str, sizeof str,  TEXT("%6.2f seconds (24 bit PCM at %s Hz)"), AvailableSpace/((float)SR[SampleGroup->ItemIndex].Value * 3.0), SR[SampleGroup->ItemIndex].Hz);
	} else {    // 4 bit ADPCM
	_sntprintf(str, sizeof str,  TEXT("%6.2f seconds (4 bit ADPCM at %s Hz)"), AvailableSpace/((float)SR[SampleGroup->ItemIndex].Value * 0.5), SR[SampleGroup->ItemIndex].Hz);
	}
MyShowMessage("Available time for new clips = " + String(str), false);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::FileSetDateNow(String FileName)
{
// Set file date and time to now
#if __BORLANDC__ >= 0x720		// CB10.1 Only
FileSetDate(FileName, DateTimeToFileDate(Now()));   // CB10
#else
int fHandle = FileOpen(FileName, fmOpenReadWrite);  // BCB 4.0
FileSetDate(fHandle, DateTimeToFileDate(Now()));
FileClose(fHandle);
#endif
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::DumpEntireCliptoDiskClick(TObject *Sender)
{
// Dump clip to disk file - exact byte by byte dump. Works for both ADPCM and PCM clips
// created file is either 4 bit IMA-ADPCM or 24 bit big endian signed PCM
//
// Dumps the clip pointed to by StringGrid1->Row. Uses ParseIndex() to set the CopyLength value correctly
//
char * membuf;
struct SFIndex DumpIndex;
DWORD readBytes, writtenBytes;
long  fp;
HANDLE hSoundFile, hRawFile;

// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clip = StringGrid1->Row;

// try and open the file
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Dump Clip: Unable to open file: " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	} else {
	MyShowMessage("Dump Clip: Sound file to be processed: " + MRUComboBox->Text, false);
	}

ParseIndex(DumpIndex, hSoundFile, clip);
// allocate memory of the correct size for a buffer - holds the entire clip
membuf = (char *) malloc(DumpIndex.ClipLength);
if (membuf == NULL) {
	MyShowMessage("Dump Clip: membuf malloc failed", true);
	CloseHandle(hSoundFile);
	return;
	}
// read from sound file from StartAddr for Length
fp = DumpIndex.StartAddr;
SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
// read length bytes
ReadFile(hSoundFile, membuf, DumpIndex.ClipLength, &readBytes, NULL);
if (readBytes != DumpIndex.ClipLength) {
	MyShowMessage("Dump Clip: Could not read clip from sound file", true);
	CloseHandle(hSoundFile);
	free(membuf);
	return;
	}
CloseHandle(hSoundFile);
// write to default named disk file
// create file
String NewExtension = "_" + FormatFloat("000", clip) + "_" + SR[SampleGroup->ItemIndex].Hz + ((DumpIndex.databyte & PCM24MASK)? "_be.raw": ".ima");
String RecordFileName = ChangeFileExt(MRUComboBox->Text, NewExtension);

// open output file
if((hRawFile = CreateFile(RecordFileName.c_str(), GENERIC_WRITE, NULL, NULL, CREATE_ALWAYS, 0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Dump Clip: Unable to open file: " + RecordFileName, true);
	RecordFlag = false;
	return;
	} else {
	MyShowMessage("Dump Clip: Dumping to file: " + RecordFileName + ": " + String(DumpIndex.ClipLength) + " bytes", false);
	}
// write to disk file
WriteFile(hRawFile, membuf, DumpIndex.ClipLength, &writtenBytes, NULL);
if (writtenBytes != DumpIndex.ClipLength)
	{
	MyShowMessage("Dump Clip: Could not write clip to .raw/.ima file", true);
	CloseHandle(hRawFile);
	free (membuf);
	return;
	}

CloseHandle(hRawFile);
free(membuf);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::DumpRepeatingSegmentofCliptoDiskClick(TObject *Sender)
{
// Dump repeating segment to disk file - exact byte by byte dump. Works for both ADPCM and PCM clips
// created file is either 4 bit IMA-ADPCM or 24 bit big endian signed PCM
//
// Dumps the clip pointed to by StringGrid1->Row. Uses ParseIndex() to set the CopyRepLength value correctly
//
char * membuf;
struct SFIndex DumpIndex;
DWORD readBytes, writtenBytes;
long  fp;
HANDLE hSoundFile, hRawFile;
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}

// try and open the file
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Dump Repeating Segment: Unable to open file: " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	} else {
	MyShowMessage("Dump Repeating Segment: Sound file to be processed: " + MRUComboBox->Text, false);
	}

const int clip = StringGrid1->Row;

ParseIndex(DumpIndex, hSoundFile, clip);

if (DumpIndex.StartRepAddr == EMPTY_FIELD) {
	MyShowMessage("Dump Repeating Segment: No repeating segment in this clip", true);
	return;   // no repeating segment there
	}

// allocate memory of the correct size for a buffer
membuf = (char *) malloc(DumpIndex.ClipLength);
if (membuf == NULL) {
	MyShowMessage("Dump Repeating Segment: membuf malloc failed", true);
	CloseHandle(hSoundFile);
	return;
	}
// read from sound file from StartRepAddr for Length
fp = DumpIndex.StartRepAddr;
SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
// read length bytes
ReadFile(hSoundFile, membuf, DumpIndex.ClipRepLength, &readBytes, NULL);
if (readBytes != DumpIndex.ClipRepLength)
	{
	MyShowMessage("Dump Repeating Segment: Could not read clip from sound file", true);
	CloseHandle(hSoundFile);
	free(membuf);
	return;
	}
CloseHandle(hSoundFile);
// write to default named disk file
// create file
String NewExtension = "_" + FormatFloat("000", clip) + "_Rep_" + SR[SampleGroup->ItemIndex].Hz + ((DumpIndex.databyte & PCM24MASK)? "_be.raw": ".ima");
String RecordFileName = ChangeFileExt(MRUComboBox->Text, NewExtension);

// open output file
if((hRawFile = CreateFile(RecordFileName.c_str(), GENERIC_WRITE, NULL, NULL, CREATE_ALWAYS, 0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Dump Repeating Segment: Unable to open file: " + RecordFileName, true);
	RecordFlag = false;
	return;
	} else {
	MyShowMessage("Dump Repeating Segment: Dumping to file: " + RecordFileName + ": " + String(DumpIndex.ClipRepLength) + " bytes", false);
	}
// write to disk file
// write length bytes
WriteFile(hRawFile, membuf, DumpIndex.ClipRepLength, &writtenBytes, NULL);
if (writtenBytes != DumpIndex.ClipRepLength) {
	MyShowMessage("Dump Repeating Segment: Could not write clip to .raw/.ima file", true);
	CloseHandle(hRawFile);
	free (membuf);
	return;
	}

CloseHandle(hRawFile);
free(membuf);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ExportAllClipstomp3Click(TObject *Sender)
{
// export all clips to mp3 files
/*
Why does LAME add silence to the beginning each song?

DECODER DELAY AT START OF FILE:

All decoders I have tested introduce a delay of 528 samples. That is, after
decoding an mp3 file, the output will have 528 samples of 0's appended to the
front. This is because the standard MDCT/filterbank routines used by the ISO
have a 528 sample delay. It would be possible to write a MDCT/filterbank routine
with a 0 sample delay (see description of Takehiro's MDCT/filterbank routine
used in LAME encoding below) but I dont know that anyone has done this.
Furthermore, because of the overlapped nature of MDCT frames, the first half of
the first granule (1 granule=576 samples) doesn't have a previous frame to
overlap with, resulting in attenuation of the first N samples. The value of N
depends on the window type. For "STOP_TYPE" and "SHORT_TYPE", N=96, while for
"START_TYPE" and "NORMAL_TYPE", N=288. The first frame produced by LAME 3.56
and up will always be of STOP_TYPE or SHORT_TYPE.

ENCODER DELAY AT START OF FILE:

ISO based encoders (BladeEnc, 8hz-mp3, etc) use a MDCT/filterbank routine
similar to the one used in decoding, and thus also introduce their own 528
sample delay. A .wav file encoded & decoded will have a 1056 sample delay
(1056 samples will be appended to the beginning).
*/
// put them into the same directory as the .mth file

if (!FileExists(String(cwd) + "\\lame.exe") || !FileExists(String(cwd) + "\\lame_enc.dll")) {
	MyShowMessage("Exporting requires that 'lame.exe' and 'lame_enc.dll' be in the ADPCM program directory. Download: http://www.rarewares.org/mp3-lame-bundle.php", true);
	return;
	}

// Loop through 255 sounds 1 to 256 (command "n1" through "n255")

Stopem = false; // press the STOP button to terminate loop early
const int LastClip = StringGrid1->RowCount;

for (int clip = 1 ; clip < LastClip ; clip++) {
	if (Stopem) {
		RecordFlag = false;	// records as 16 bit little endian signed PCM
		return;		// exit
		}
	ExportClip(clip, false);  // Export
	}
MyShowMessage("Export All Clips: Complete", true);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ExportAllClipsforDFPlayer1Click(TObject *Sender)
{
// put them into the same directory as the .mth file

if (!FileExists(String(cwd) + "\\lame.exe") || !FileExists(String(cwd) + "\\lame_enc.dll")) {
	MyShowMessage("Exporting requires that 'lame.exe' and 'lame_enc.dll' be in the ADPCM program directory. Download: http://www.rarewares.org/mp3-lame-bundle.php", true);
	return;
	}

// Loop through 255 sounds 1 to 256 (command "n1" through "n255")

Stopem = false; // press the STOP button to terminate loop early
const int LastClip = StringGrid1->RowCount;

for (int clip = 1 ; clip < LastClip ; clip++) {
	if (Stopem) {
		RecordFlag = false;	// records as 16 bit little endian signed PCM
		return;		// exit
		}
	ExportClip(clip, true);  // Export with filename compatible with DFPlayer
	}
MyShowMessage("Export All Clips: Complete", true);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1MouseDown(TObject *Sender, TMouseButton Button,
          TShiftState Shift, int X, int Y)
{
if (!Decoded) return;
int Column, Row;
StringGrid1->MouseToCell(X, Y, Column, Row);
//MyShowMessage("Mouse Down in index # " + String(Row), false);
if (Button == mbLeft) { // drag only if left button pressed
	if (Row == 0) return; // ignore Mouse Down in fixed rows
#ifdef DRAGDROP
    // see note at top of file
	TStringGrid *pSG = (TStringGrid *)Sender; // cast to TStringGrid
	if (Column == 0) {  // are we in Column 0 ?
		pSG->BeginDrag(false,-1);                   // if so, drag it
		}
#endif
	return; // ignore mouse down in all data cells
	}
if (Button == mbRight) {
	// cause row under mouse to be selected
    StringGrid1->Row = Row;
	IndexPopupMenu->Popup(Left + X, Top + Y); // display popup menu
	return;
	}
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1DragOver(TObject *Sender, TObject *Source,
          int X, int Y, TDragState State, bool &Accept)
{
if (IsDragObject(Source)) {
	// are we in Column 0 ?
	int Column, Row;
	StringGrid1->MouseToCell(X, Y, Column, Row);
	if (Row == 0) {
		Accept = false;
		return; // ignore drag over fixed rows
		}
	if (Column == 0) {
		Accept = true;                   // if so, accept it
		return;
		}
	}
Accept = false;
// ignore drag over all data cells
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1DragDrop(TObject *Sender, TObject *Source,
          int X, int Y)
{
TMyDragObject * SG;
SG = dynamic_cast<TMyDragObject *>(Source);
//
if (IsDragObject(Source)) {
	// are we in Column 0 ?
	int Column, Row;
	StringGrid1->MouseToCell(X, Y, Column, Row);
	if (Row == 0) {
		return; // ignore drop over fixed rows
		}
	if (Column == 0) {	// if so, write the draged index entry into this index entry on disk
		if (SG != NULL) {   // open the file, write the index, close the file, call AnalyzeButtonClick()
			// dump out the Drag Object to the sound file on disk
			HANDLE hIndexFile;/* file handle */
			/*
			 * try and open the file
			 */
			if((hIndexFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_WRITE, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
				MyShowLastError("Drag Drop: Unable to open file: " + MRUComboBox->Text, true);
				return;
				} else {
				MyShowMessage("Drag Drop: Dropped clip " + String(SG->SourceClipNo) + " onto clip " + String(Row) + " in file: " + MRUComboBox->Text, false);
				}
			// write a 16 byte index entry
			// clip number is 1 based
			SetFilePointer(hIndexFile, 0x0100 + ((Row - 1) * 16), NULL, FILE_BEGIN);
			DWORD writtenBytes;
			// write 16 bytes from the DragObject
			WriteFile(hIndexFile, SG->IndexSource, sizeof(SG->IndexSource), &writtenBytes, NULL);
			if (writtenBytes != sizeof(SG->IndexSource)) {
				ADPCM_Main->MyShowMessage("Drag Drop: Write did not write 16 bytes", true);
				CloseHandle(hIndexFile);
				return;
				}
			CloseHandle(hIndexFile);
			// call to AnalyzeButtonClick();
			AnalyzeButtonClick(Sender);
			}
		return;
		}
	}
// ignore drop in all data cells
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1EndDrag(TObject *Sender, TObject *Target,
          int X, int Y)
{
if (Target) {
	// drop was successful but I don't care
    }
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1StartDrag(TObject *Sender, TDragObject *&DragObject)
{
TPoint pt;
GetCursorPos(&pt);  // current mouse position
int Column, Row;
pt = StringGrid1->ScreenToClient(pt);
StringGrid1->MouseToCell(pt.x, pt.y, Column, Row);
//
TStringGrid * SG;
SG = dynamic_cast<TStringGrid *>(Sender);
//

if (SG != NULL) {
	// load up the Drag Object from the sound file on disk
	HANDLE hIndexFile;/* file handle */
	/*
	 * try and open the file
	 */
	if((hIndexFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
		MyShowLastError("Start Drag: Unable to open file: " + MRUComboBox->Text, true);
		return;
		} else {
		MyShowMessage("Start Drag: Dragging clip " + String(Row) + " of file: " + MRUComboBox->Text, false);
		}
	// get a 16 byte index entry
	// clip number is 1 based
	SetFilePointer(hIndexFile, 0x0100 + ((Row - 1) * 16), NULL, FILE_BEGIN);
	DWORD readBytes;
	// read 16 bytes into the DragObject
	ReadFile(hIndexFile, tm->IndexSource, sizeof(tm->IndexSource), &readBytes, NULL);
	if (readBytes != sizeof(tm->IndexSource)) {
		ADPCM_Main->MyShowMessage("Start Drag: Read did not read 16 bytes", true);
		DragObject = NULL;  // fail
		return;
		}
	CloseHandle(hIndexFile);
    tm->SourceClipNo = Row;
	tm->SourceFilename = MRUComboBox->Text;
	}
DragObject = tm;
}
//---------------------------------------------------------------------------

__fastcall TMyDragObject::TMyDragObject(void)
{
TDragObject();
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::EditOneByte(TObject *Sender)
{
// NOT USED Save this code for later use ----------------------
// Must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int clip = StringGrid1->Row;

// read the sound file index entry
HANDLE hSoundFile;
SFIndex IndexSound;
// try and open the file
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("EditOneByte: Unable to open file: " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	} else {
	MyShowMessage("EditOneByte: Sound file to be edited: " + MRUComboBox->Text, false);
	}
// read one byte
SetFilePointer(hSoundFile, 0x0100 + ((clip - 1) * 16) + 8, NULL, FILE_BEGIN);
DWORD readBytes;
byte bytetoedit[1];
// read 1 byte
ReadFile(hSoundFile, bytetoedit, 1, &readBytes, NULL);
if (readBytes != 1) {
	ADPCM_Main->MyShowMessage("EditOneByte: Read did not read 1 byte", true);
	return;
	}

if (bytetoedit[0] == 0xFF) return;  // nothing to do, no Start Repeat Address

MyShowMessage("EditOneByte: Edit One Byte Index #" + String(clip), false);

// make it point to the odd nybble
bytetoedit[0] |= 0x01;

// write one byte
SetFilePointer(hSoundFile, 0x0100 + ((clip - 1) * 16) + 8, NULL, FILE_BEGIN);
DWORD byteswritten;
WriteFile(hSoundFile, bytetoedit, 1, &byteswritten, NULL);
if (byteswritten != 1) {
	ADPCM_Main->MyShowMessage("EditOneByte: Write did not write 1 byte", true);
	return;
	}

CloseHandle(hSoundFile);
AnalyzeButtonClick(Sender);
}
//---------------------------------------------------------------------------

bool ParseQuillIndex(QIndex &Index, HANDLE hParseFile, int index)
{
// get a 16 byte quill index entry and parse it into the QIndex structure named Index
// Index number is 1 based
SetFilePointer(hParseFile, QUILL_INDEX + ((index - 1) * 16), NULL, FILE_BEGIN);
unsigned char buffer16[16];
DWORD readBytes;
// read 16 bytes
ReadFile(hParseFile, buffer16, sizeof(buffer16), &readBytes, NULL);
if (readBytes != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("Parse Quill Index: Read did not read 16 bytes", true);
	return false;
	}
//
Index.dataword =  (buffer16[0] << 24) |(buffer16[1] << 16) | (buffer16[2] << 8) | buffer16[3];    // bytes 0-3
Index.StartAddr =  (buffer16[4] << 16) | (buffer16[5] << 8) | buffer16[6];    // bytes 4-5
Index.B7 = buffer16[7];	// always 0x00
Index.EndAddr =   (buffer16[8] << 16) | (buffer16[9] << 8) | buffer16[10];    // bytes 8-10
Index.B11 = buffer16[11];   // always 0x80
Index.B12 = buffer16[12];	// always 0x00
Index.SMInit =   (buffer16[13] << 16) | (buffer16[14] << 8) | buffer16[15];    // bytes 13-15
return true;
}
//---------------------------------------------------------------------------

bool WriteQuillIndex(QIndex &Index, HANDLE hParseFile, int index)
{
// write a 16 byte index entry from the QIndex structure named Index
// index number is 1 based
SetFilePointer(hParseFile, QUILL_INDEX + ((index - 1) * 16), NULL, FILE_BEGIN);
unsigned char buffer16[16];

buffer16[0] = (Index.dataword & 0xFF000000) >> 24;
buffer16[1] = (Index.dataword & 0x00FF0000) >> 16;
buffer16[2] = (Index.dataword & 0x0000FF00) >> 8;
buffer16[3] = (Index.dataword & 0x000000FF);
buffer16[4] = (Index.StartAddr & 0xFF0000) >> 16;
buffer16[5] = (Index.StartAddr & 0x00FF00) >> 8;
buffer16[6] = (Index.StartAddr & 0x0000FF);
buffer16[7] = Index.B7; // always 0x00
buffer16[8] =  (Index.EndAddr & 0xFF0000) >> 16;
buffer16[9] =  (Index.EndAddr & 0x00FF00) >> 8;
buffer16[10] = (Index.EndAddr & 0x0000FF);
buffer16[11] = Index.B11;   // always 0x80
buffer16[12] = Index.B12;	// always 0x00
buffer16[13] = (Index.SMInit & 0xFF0000) >> 16;
buffer16[14] = (Index.SMInit & 0x00FF00) >> 8;
buffer16[15] = (Index.SMInit & 0x0000FF);
//
DWORD byteswritten;
WriteFile(hParseFile, buffer16, sizeof(buffer16), &byteswritten, NULL);
if (byteswritten != sizeof(buffer16)) {
	ADPCM_Main->MyShowMessage("Write Quill Index: Write did not write 16 bytes", true);
	return false;
	}
return true;
}
//---------------------------------------------------------------------------
void __fastcall TADPCM_Main::QuillButtonClick(TObject *Sender)
{
// Show the Quilling Whistle Data Index
// Must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
const int DataVal = _tcstol(StringGrid1->Cells[IDATABYTE][StringGrid1->Row].c_str(), NULL, 16);
if ((((DataVal & 0x30) >> 4) != HornWhistleSound) && (((DataVal & 0x30) >> 4) != AccentSound)) {
	MyShowMessage("You must select a Whistle clip (and it probably should be clip #3)", true);
	return;
	}
const int clip = StringGrid1->Row;  // I believe this is always #3
//
if (Quill_Index == NULL) {
	Quill_Index = new TQuill_Index(Application, MRUComboBox->Text, clip);
	assert(Quill_Index != NULL);
	}

Quill_Index->ShowIt(MRUComboBox->Text, clip);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::StringGrid1MouseMove(TObject *Sender, TShiftState Shift,
          int X, int Y)
{
int R, C;
static int Row = -1 , Col = -1; // hold previous row and col so we don't regenerate the hint over and over
// Must have analyzed a sound file
if (!Decoded) {
	return;
	}
StringGrid1->MouseToCell(X, Y, C, R);
//
// Use this to make a custom hint for each cell of the StringGrid
if (Row != R || Col != C) {
	Row = R;
	Col = C;
	Application->CancelHint();  // take down any hint that might be displayed
	if (Row == 0) {
		if (Col == ICLIPNO) {
			StringGrid1->Hint = "Click to toggle column between clip # and offset";
			return;
			}
		if (Col == ISTARTADDR) {
			StringGrid1->Hint = "Clip Start Address";
			return;
			}
		if (Col == IENDREPADDR) {
			StringGrid1->Hint = "Repeating Segment End Address (Red or Green indicates PCM address is not modulo 3)";
			return;
			}
		if (Col == ISTARTREPADDR) {
			StringGrid1->Hint = "Repeating Segment Start Address (Red or Green indicates PCM address is not modulo 3)";
			return;
			}
		if (Col == ISMINIT) {
			StringGrid1->Hint = "ADPCM State Machine Initialization";
			return;
			}
		if (Col == IENDADDR) {
			StringGrid1->Hint = "Clip End Address (Red or Green indicates PCM address is not modulo 3)";
			return;
			}
		if (Col == ILENGTH) {
			StringGrid1->Hint = "Calculated Clip Length (decimal) Red or Green indicates PCM length is not modulo 3";
			return;
			}
		if (Col == IDATABYTE) {                      
			StringGrid1->Hint = "Clip type and format";
			return;
			}
		}
	if (Row > 0 && Col == 0) {
		// see note at top of file
		//StringGrid1->Hint = "Drag to copy index to another location";
		byte databyte = _tcstol(StringGrid1->Cells[IDATABYTE][Row].c_str(), NULL, 16);
		if (StringGrid1->Cells[ISTARTADDR][Row] == "") 
			StringGrid1->Hint = "Index row is empty";
			else
			if (databyte & PCM24MASK)
				StringGrid1->Hint = "Clip is 24 bit signed PCM";
				else
				StringGrid1->Hint = "Clip is 4 bit ADPCM";
		return;
		}
	if (Row > 0) {

		byte databyte = _tcstol(StringGrid1->Cells[IDATABYTE][Row].c_str(), NULL, 16);
		if (databyte  & PCM24MASK) {   // 24 pit PCM hints
			if ((Col == IENDREPADDR) || (Col == ISTARTREPADDR) || (Col == IENDADDR)) {
				StringGrid1->Hint = "Red or Green indicates address is not modulo 3";
				return;
				}
			if (Col == ISMINIT) {
				// TODO display value of field as a 16 bit signed int
				//StringGrid1->Hint = "Purpose of this field for PCM format is not known";
				int32_t iVal = _tcstol(StringGrid1->Cells[ISMINIT][Row].c_str(), NULL, 16);
				int16_t sVal = iVal;	// TBD was:  & 0x0000FFFF;
				//iVal = sVal;
				StringGrid1->Hint = "Value = " + String(sVal);
				return;
			}
			if (Col == ILENGTH) {
				StringGrid1->Hint = "Clip length in bytes (" +  String(_tcstol(StringGrid1->Cells[ILENGTH][Row].c_str(), NULL, 10)/3) + " samples)";
				return;
				}
			if (Col == IDATABYTE) {
				StringGrid1->Hint = "Color indicates clip type";
				return;
				}
			} else {
			if (databyte != 0x00) {
			// 4 bit ADPCM hints
				if (Col == ISMINIT) {
					StringGrid1->Hint = "ADPCM State Machine Initialization for the Repeating Segment";
					return;
					}
				//if ((Col == IENDREPADDR) || (Col == ISTARTREPADDR) || (Col == IENDADDR)) {
				//	StringGrid1->Hint = "Green indicates address is an odd 4 bit ADPCM nybble";
				//	return;
				//	}
				if (Col == ILENGTH) {
					StringGrid1->Hint = "Clip length in bytes (" +  String(_tcstol(StringGrid1->Cells[ILENGTH][Row].c_str(), NULL, 10)*2) + " samples)";
					return;
					}
				if (Col == IDATABYTE) {
					StringGrid1->Hint = "Color indicates clip type";
					return;
					}
				}
			}
		StringGrid1->Hint = "Clip # " + IntToStr(R);	// new hint
		}
	}

}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::RemoveFromListClick(TObject *Sender)
{
MRUComboBox->Items->Delete(MRUComboBox->ItemIndex);
MRUComboBox->ItemIndex = -1;	// nothing selected	
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::TruncateSoundFileforFasterDownloadingClick(TObject *Sender)

{
#define SIZE 0x8000
#define cSIZE "0x8000"
// truncate sound file at SIZE for faster downloading
// Must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
// read the sound file index entry
HANDLE hSoundFile;
SFIndex IndexSound;
DWORD readBytes, writtenBytes;
// try and open the file
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Truncate: Unable to open file: " + MRUComboBox->Text, true);
	// cancel the operation
	return;
	} else {
	MyShowMessage("Truncate: Sound file to be truncated: " + MRUComboBox->Text, false);
	}

// Allocate 0x8000 bytes
char * membuf;
membuf = (char *) malloc(SIZE);

// read bytes
ReadFile(hSoundFile, membuf, SIZE, &readBytes, NULL);
if (readBytes != SIZE) {
	MyShowMessage("Truncate: Could not read from Sound file", true);
	CloseHandle(hSoundFile);
	free(membuf);
	return;
    }
CloseHandle(hSoundFile);

// create new filename for truncated Sound file
String NewExtension =  String("_") + cSIZE + String(".mth");
const String TruncatedFileName = ChangeFileExt(MRUComboBox->Text, NewExtension);

// delete file if present
if (FileExists(TruncatedFileName))
	DeleteFile(TruncatedFileName);

// open output file
if((hSoundFile = CreateFile(TruncatedFileName.c_str(), GENERIC_WRITE, NULL, NULL, CREATE_ALWAYS, 0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Truncate: Unable to open file: " + TruncatedFileName, true);
	return;
	} else {
	MyShowMessage("Truncate: Exporting Truncated Sound file to file: " + TruncatedFileName, false);
	}


// write SIZE bytes
WriteFile(hSoundFile, membuf, SIZE, &writtenBytes, NULL);
if (writtenBytes != SIZE) {
	ADPCM_Main->MyShowMessage("Truncate: Could not write to Sound file", true);
	free (membuf);
	CloseHandle(hSoundFile);
	return;
	}
// Close
free (membuf);
CloseHandle(hSoundFile);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::MRUComboBoxChange(TObject *Sender)
{
Decoded = false;
ClearButtonClick(Sender);
if (Edit_Engine_Parameters) {
	delete Edit_Engine_Parameters;
	Edit_Engine_Parameters = NULL;
	}
if (Edit_Scale_Factor) {
	delete Edit_Scale_Factor;
	Edit_Scale_Factor = NULL;
	}
MRUComboBox->Hint = MRUComboBox->Text;
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::ChangePCMClipVolume1Click(TObject *Sender)
{
// Change the volume of a PCM clip
// Must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
int ARow = StringGrid1->Row;
if (!((int) StringGrid1->Objects[IDATABYTE][ARow] & PCM24MASK)) {
	// Not a PCM clip
	MyShowMessage("Volume: This clip is not a PCM clip", true);
	return;
	}

const String Mult = InputBox("Change Clip Volume", "Enter decimal value", "1.0");
const double MV = _tcstod(Mult.c_str(), NULL);
MyShowMessage("Volume: Multiplier = " + String(MV), false);

// Use _tcstol for both BCB4 and CB10          see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
const int StartAddr = _tcstol(StringGrid1->Cells[ISTARTADDR][ARow].c_str(), NULL, 16);
const int EndAddr = _tcstol(StringGrid1->Cells[IENDADDR][ARow].c_str(), NULL, 16);
//const int DataVal = _tcstol(StringGrid1->Cells[IDATABYTE][StringGrid1->Row].c_str(), NULL, 16);

MyShowMessage("Volume: Clip #" + String(ARow) + " Starting Address " + StringGrid1->Cells[ISTARTADDR][ARow] + " Ending Address " + StringGrid1->Cells[IENDADDR][ARow], false);

// read in the clip
/*
 * try and open the file
 */
HANDLE hSoundFile;
if((hSoundFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
	MyShowLastError("Volume: Unable to open file: " + MRUComboBox->Text, true);
	return;
	} else {
	MyShowMessage("Volume: Processing file: " + MRUComboBox->Text, false);
	}

// go through 3 bytes a at time and raise the volume
if (EndAddr != -1 && StartAddr > EndAddr) {
	ADPCM_Main->MyShowMessage("Volume:StartAddr > EndAddr", true);
	return;
	}

unsigned char cBuffer3[3];	// holds 2 4 bit ADPCM Samples or 1 24 bit big endian signed PCM Sample

// set file pointer to first byte of the seqment
long  fp = StartAddr;
String MAS;

MAS = HEXFIELDPREFIX + IntToHex((int)StartAddr, 6);
ADPCM_Main->MyShowMessage("Volume: PCM Clip at Enter (" + MAS + ")", false);

// set length of seqment
const int len = EndAddr - StartAddr;

DWORD readBytes, writeBytes;
int TotalBytes = len;
int iBCount = 0;
bool Done = false;

int AudioMin24 = 8388607;
int AudioMax24 = -8388608;

//
while(1) {
	int Value1;
	if (Done)	break;	// done
	//---
	switch(ADPCM_Main->FormatGroup->ItemIndex) {
        default:
		case ADPCM:		// 4 bit ADPCM converted to 16 bit signed PCM
			// TODO implement 4 bit ADPCM volume change
			return;
		case PCM24:		// 24 bit signed PCM converted to 16 bit signed PCM
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			if(!ReadFile(hSoundFile, cBuffer3, 3, &readBytes, NULL)) break;	// read 24 bits
			if(readBytes != 3) {
				Done = true;
				break;	// out of for loop - need three bytes to process
				}
			TotalBytes -= readBytes;	// subtract bytes read if it goes negative, we are done
			// Big Endian
			Value1 = ((cBuffer3[0] << 16) | (cBuffer3[1] << 8) | cBuffer3[2]);	// 24 bit big endian to 24 bit signed PCM
			if (Value1 & 0x00800000) Value1 |= 0xFF000000;	// convert to 32 bit signed PCM

			Value1 *= MV;	// change the volume
			if (Value1 > AudioMax24) {
            	AudioMax24 = Value1;	// retain min and max audio value for debugging
                }
			if (Value1 < AudioMin24) {
            	AudioMin24 = Value1;
                }
			#ifdef DEBUGONLY
			// Debug - show each 24 bit sound value
			if (iBCount < 20 || iBCount > 47300) ADPCM_Main->MyShowMessage(String(iBCount) + " " + HEXFIELDPREFIX + IntToHex((int)(Value1 & 0xFFFFFFFF), 8) +
				" (" + String(Value1) + ")", false);
			#endif

			// convert back to 24 bit big endian
			cBuffer3[0] = (Value1 & 0x00FF0000) >> 16;
			cBuffer3[1] = (Value1 & 0x0000FF00) >> 8;
			cBuffer3[2] = (Value1 & 0x000000FF);

			// write out the sample
			SetFilePointer(hSoundFile, fp, NULL, FILE_BEGIN);
			if(!WriteFile(hSoundFile, cBuffer3, 3, &writeBytes, NULL)) break;	// write 24 bits
			if(writeBytes != 3) {
				Done = true;
				break;	// out of for loop - unknown failure
				}
			iBCount += 3;	// count bytes read from sound file
			fp += 3;		// increment the file pointer
			if (TotalBytes <= 0) {	// all done after this buffer is processed
				Done = true;
				break;
				}
			break;
		case PCM16: // 16 bit signed PCM no conversion needed
			return;
		}
	if(Done) break;
	//---
	Application->ProcessMessages();
	if (Stopem) break;	// cancel
	}
MAS = HEXFIELDPREFIX + IntToHex((int)StartAddr + iBCount, 6);
ADPCM_Main->MyShowMessage("Volume: PCM Clip at Exit (" + MAS + ")", false);
TCHAR str[25], str2[25];
_sntprintf(str, sizeof str, TEXT("%6.2f"), (AudioMax * 100)/(8388607.0));
_sntprintf(str2, sizeof str, TEXT("%6.2f"), (AudioMin * 100)/(8388607.0));
MyShowMessage("Volume : 16 bit Dynamic Range = " + String(AudioMax/256) + " to " + String(AudioMin/256) + " (" + String(str) +  "% to " + String(str2) + "%)", false);
_sntprintf(str, sizeof str, TEXT("%6.2f"), (AudioMax * 100)/(8388607.0));
_sntprintf(str2, sizeof str, TEXT("%6.2f"), (AudioMin * 100)/(8388607.0));
MyShowMessage("Volume : 24 bit Dynamic Range = " + String(AudioMax) + " to " + String(AudioMin) + " (" + String(str) +  "% to " + String(str2) + "%)", false);

CloseHandle(hSoundFile);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::EdittheScaleFactor1Click(TObject *Sender)
{
// edit the Scale Factor
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
//
if (Edit_Scale_Factor == NULL) {
	Edit_Scale_Factor = new TEdit_Scale_Factor(Application, MRUComboBox->Text, 1, 1, PSGroup->ItemIndex);
	assert(Edit_Scale_Factor != NULL);
	}

Edit_Scale_Factor->ShowIt(MRUComboBox->Text, 1, 1, PSGroup->ItemIndex);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::CleartheList1Click(TObject *Sender)
{
// clear the list
MRUComboBox->Text = "";
MRUComboBox->Items->Clear();
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::HiddenDblClick(TObject *Sender)
{
iDebug++;
Hidden->Hint = "Debug = " + String(iDebug);
MyShowMessage(Hidden->Hint, false);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::PlayClipButtonMouseEnter(TObject *Sender)
{
if (!Decoded) {
	PlayClipButton->Hint = "Play the hightlighted clip";
	return;
	}
// create hint with clip time
const int clip = StringGrid1->Row;

float Secs;

// Use _tcstol  for both BCB4 and CB10           see http://docwiki.embarcadero.com/RADStudio/Berlin/en/Floating_Functions
if (StringGrid1->Cells[ISTARTADDR][clip] != "") {
	const int Length =  StringGrid1->Cells[ILENGTH][clip].ToInt();    // Length base 10
	const int DataVal = _tcstol(StringGrid1->Cells[IDATABYTE][clip].c_str(), NULL, 16);
	const int nSamplesPerSec = SR[SampleGroup->ItemIndex].Value;
	if (DataVal & PCM24MASK)
		// PCM24
		Secs = (Length / 3.0) / nSamplesPerSec;
		else
		// ADPCM
		Secs = (Length * 2.0) / nSamplesPerSec;
	PlayClipButton->Hint = "Play the hightlighted clip (" + FormatFloat(".00", Secs) + " seconds)";	
	}
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::EditEngineParameters1Click(TObject *Sender)
{
// edit the Engine Parameters
// must have analyzed a sound file
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
//
if (Edit_Engine_Parameters == NULL) {
	Edit_Engine_Parameters = new TEdit_Engine_Parameters(Application, MRUComboBox->Text, 1, 1, PSGroup->ItemIndex);
	assert(Edit_Engine_Parameters != NULL);
	}

Edit_Engine_Parameters->ShowIt(MRUComboBox->Text, 1, 1, PSGroup->ItemIndex);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::EditParamsButtonClick(TObject *Sender)
{
// edit the Engine Parameters
// must have analyzed a sound file
// Sames as Menu item in ADPCMPopupMenu
if (!Decoded) {
	MyShowMessage("Press the Analyze button first", true);
	return;
	}
//
if (Edit_Engine_Parameters == NULL) {
	Edit_Engine_Parameters = new TEdit_Engine_Parameters(Application, MRUComboBox->Text, 1, 1, PSGroup->ItemIndex);
	assert(Edit_Engine_Parameters != NULL);
	}

Edit_Engine_Parameters->ShowIt(MRUComboBox->Text, 1, 1, PSGroup->ItemIndex);
}
//---------------------------------------------------------------------------

void __fastcall TADPCM_Main::Copylower256bytesClick(TObject *Sender)
{
//
// select source sound file
OpenDialog->DefaultExt = "mth";
OpenDialog->Filter = "Sound files (*.mth)|*.MTH";
OpenDialog->Options.Clear();
OpenDialog->Options << ofFileMustExist;
OpenDialog->Title = "Select Sound File to Copy Lower 256 Bytes From";

if (OpenDialog->Execute()) {
	// file picked
    HANDLE hCopyFile, hDestFile;
	// try and open the source file
    if((hCopyFile = CreateFile(OpenDialog->FileName.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
        MyShowLastError("Copy 256 bytes: Unable to open file: " + OpenDialog->FileName, true);
        return;
        } else {
        MyShowMessage("Copying 256 bytes from : " + OpenDialog->FileName, false);
        }
    // read in 256 bytes starting at offset 0x00
    (void) SetFilePointer(hCopyFile, 0, NULL, FILE_BEGIN);
	unsigned char InternalString[256];
	DWORD readBytes;
	int retv = ReadFile(hCopyFile, InternalString, 256, &readBytes, NULL);
	if (retv != true || readBytes != 256) {
		MyShowMessage("getFileString: didn't read 256 bytes", false);
        CloseHandle(hCopyFile);
        return;
		}
    // close the file
	CloseHandle(hCopyFile);
	// write out 256 bytes to current sound file
	// try and open the destination file
	if((hDestFile = CreateFile(MRUComboBox->Text.c_str(), FILE_ALL_ACCESS, NULL, NULL, OPEN_EXISTING,0, NULL)) == INVALID_HANDLE_VALUE) {
        MyShowLastError("Copy 256 bytes: Unable to open file: " + MRUComboBox->Text, true);
        return;
        } else {
        MyShowMessage("Copying 256 bytes to : " + MRUComboBox->Text, false);
        }
    // write out 256 bytes starting at offset 0x00
    (void) SetFilePointer(hDestFile, 0, NULL, FILE_BEGIN);
    DWORD writeBytes;
	retv = WriteFile(hDestFile, InternalString, 256, &writeBytes, NULL);
	if (retv != true || writeBytes != 256) {
		MyShowMessage("getFileString: didn't write 256 bytes", false);
        CloseHandle(hDestFile);
        return;
		}
    // close the file
	CloseHandle(hDestFile);
    // reload the Engine Parameters
    if (Edit_Engine_Parameters) {
        Edit_Engine_Parameters->ReloadButtonClick(NULL);
    	}
    MyShowMessage("Done!", false);
	}
}
//---------------------------------------------------------------------------

