// vox2wav.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include ".\vox2wav.h"


CVox2Wav::CVox2Wav(void)
{
	m_szLastError = NULL;

	m_nLastSample = 0;
	m_nCurrentStep = 0;

	//populate the vaw header
	m_sWavHeader.pcmfmt.wf.nChannels = 1;	//save only mono
	//the rest of the fields will be completed when Convert is called
}

CVox2Wav::~CVox2Wav(void)
{
	free(m_szLastError);
}

BOOL CVox2Wav::Convert(LPCTSTR szVoxFileName, LPCTSTR szWavFileName, int nSamplesPerSec, int nWavBitsPerSample)
{
	//free the last error, maybe the object is reused
	free(m_szLastError);

	//check the parameters
	if(nSamplesPerSec != 6000 && nSamplesPerSec != 8000)
	{
		//Dialogic only supports 6000 and 8000 vox
		m_szLastError = _tcsdup(_T("Only 6000 or 8000 samples per second are supported!"));
		return FALSE;
	}
	if(nWavBitsPerSample != 8 && nWavBitsPerSample != 16)
	{
		//we support only 8 or 16 bits per sample
		m_szLastError = _tcsdup(_T("Only 8 or 16 bits per sample are supported!"));
		return FALSE;
	}
	if((lstrlen(szVoxFileName) == 0) || (lstrlen(szWavFileName) == 0))
	{
		//we need the vox path
		m_szLastError = _tcsdup(_T("Vox path or Wav path is empty!"));
		return FALSE;
	}
	
	//check to see if we can open the files
	HANDLE hVOX = CreateFile(szVoxFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if(hVOX == INVALID_HANDLE_VALUE)
	{
		//the vox file cannot be created!
		m_szLastError = _tcsdup(_T("Vox file cannot be opened!"));
		return FALSE;
	}
	HANDLE hWAV = CreateFile(szWavFileName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if(hWAV == INVALID_HANDLE_VALUE)
	{
		//the vox file cannot be created!
		m_szLastError = _tcsdup(_T("Vav file cannot be created!"));
		CloseHandle(hVOX);
		return FALSE;
	}

	//now update the wav header and write it into the wav file
	int nVoxBytes = GetFileSize(hVOX, NULL);
	//assume that the GetFileSize does not fail if the file was opened successfully
	int nSamples = nVoxBytes * 2;	//each VOX byte has 2 samples!
	int nWavRawBytes = nSamples;
	if(nWavBitsPerSample == 16)
		nWavRawBytes *= 2;
	int nWavFileLen = nWavRawBytes + sizeof(MDT_PCM_WAV_HEADER);
	m_sWavHeader.dwChunkSize = nWavFileLen - 8;
	m_sWavHeader.pcmfmt.wf.nSamplesPerSec = nSamplesPerSec;
	m_sWavHeader.pcmfmt.wf.nAvgBytesPerSec = nSamplesPerSec * nWavBitsPerSample/8;	//it's actually SampleRate * NumChannels * BitsPerSample/8, but NumChannels is always 1
	m_sWavHeader.pcmfmt.wf.nBlockAlign = nWavBitsPerSample/8;	//it's actually NumChannels * BitsPerSample/8, but NumChannels is always 1
	m_sWavHeader.pcmfmt.wBitsPerSample = nWavBitsPerSample;
	m_sWavHeader.dataSize = nWavRawBytes;
	//write into the wav
	DWORD dwWritten = 0;
	if(WriteFile(hWAV, &m_sWavHeader, sizeof(MDT_PCM_WAV_HEADER), &dwWritten, NULL) == FALSE)
	{
		//cannot write to wav!
		m_szLastError = _tcsdup(_T("Cannot write to the wav file!"));
		CloseHandle(hVOX);
		CloseHandle(hWAV);
		DeleteFile(szWavFileName);
		return FALSE;
	}

	//start converting
	//we need a read buffer and a write buffer
	BYTE buffRead[VOX_DECODE_BUFFER_SIZE];	//read 8 K at a time
	short buffWrite[VOX_DECODE_BUFFER_SIZE*2];	//each vox byte has 2 samples which are decoded to a 12 bit value
	DWORD dwRead = VOX_DECODE_BUFFER_SIZE;
	while(dwRead == VOX_DECODE_BUFFER_SIZE)
	{
		if(ReadFile(hVOX, buffRead, VOX_DECODE_BUFFER_SIZE, &dwRead, NULL) == FALSE)
		{
			//cannot read from vox!
			m_szLastError = _tcsdup(_T("Cannot read from the vox file!"));
			CloseHandle(hVOX);
			CloseHandle(hWAV);
			DeleteFile(szWavFileName);
			return FALSE;
		}

		//process each byte
		int idxWrite = 0;
		for(DWORD idxRead = 0; idxRead < dwRead; idxRead++)
		{
			buffWrite[idxWrite++] = Decode((buffRead[idxRead]>>4)&0x0f);
			buffWrite[idxWrite++] = Decode(buffRead[idxRead]&0x0f);
		}

		//write the current decoded samples
		if(!WriteDecodedSamples(hWAV, nWavBitsPerSample, buffWrite, dwRead*2))
		{
			//cannot write to wav!
			m_szLastError = _tcsdup(_T("Cannot write to the wav file!"));
			CloseHandle(hVOX);
			CloseHandle(hWAV);
			DeleteFile(szWavFileName);
			return FALSE;
		}
	}

	//clear everything and exit
	CloseHandle(hVOX);
	CloseHandle(hWAV);

	return TRUE;
}

void CVox2Wav::GetLastError(LPTSTR szErrorBuff, int nErrorBuffLen)
{
	if(szErrorBuff == NULL || nErrorBuffLen < 1)
		return;

	if(m_szLastError == NULL)
	{
		szErrorBuff[0] = _T('\0');
		return;
	}

	lstrcpyn(szErrorBuff, m_szLastError, nErrorBuffLen-1);
	szErrorBuff[nErrorBuffLen-1] = _T('\0');
}

short CVox2Wav::Decode(BYTE nSample)
{
	//we should get only one nibble
	ASSERT((nSample & 0x0f) == nSample);

	short StepSize = VOX_STEP_SIZE[m_nCurrentStep];
	short Offset = StepSize/8;
	if (nSample & 0x01)
		Offset += StepSize/4;
	if (nSample & 0x02)
		Offset += StepSize/2;
	if (nSample & 0x04)
		Offset += StepSize;
	short diff = (nSample & 0x08) ? -Offset : Offset;
	short nRet = m_nLastSample + diff;

	//adjust the nRet to be 12 bits
	if(nRet > 2047)
		nRet = 2047;
	if(nRet < -2048)
		nRet = -2048;

	//calculate the next step
	m_nLastSample = nRet;
	m_nCurrentStep += VOX_STEP_ADJUST[nSample&0x07];
	if(m_nCurrentStep < 0)
		m_nCurrentStep = 0;
	if(m_nCurrentStep > VOX_NR_STEPS-1)
		m_nCurrentStep = VOX_NR_STEPS-1;
	
	return nRet;
}

BOOL CVox2Wav::WriteDecodedSamples(HANDLE hWav, int nWavBitsPerSample, short* pSamples, int nrSamples)
{
	if(nWavBitsPerSample == 16)
	{
		//it's simpler... just convert in place the 12 bit to 16 bit and write the buffer
		for(int idx = 0; idx < nrSamples; idx++)
			pSamples[idx] *= 16;	//just move to the left 4 bits!
		//write the buffer
		DWORD dwWritten = 0;
		if(WriteFile(hWav, pSamples, nrSamples*sizeof(short), &dwWritten, NULL) == FALSE)
			return FALSE;
		if(dwWritten != nrSamples*sizeof(short))
			return FALSE;
		return TRUE;
	}
	else if(nWavBitsPerSample == 8)
	{
		//we have to allocate a 8 bit buffer in which to transform the samples from 12 to 8 bit
		//we know that the number of samples is maximum VOX_DECODE_BUFFER_SIZE*2
		ASSERT(nrSamples <= VOX_DECODE_BUFFER_SIZE*2);
		BYTE buff8bit[VOX_DECODE_BUFFER_SIZE*2];
		for(int idx = 0; idx < nrSamples; idx++)
		{
			//shift the sample value right 4 bits
			//but add 128 because stupid 8 bit wav is not from -128 to 127, but from 0 to 255
			pSamples[idx] /= 32;	//why not 16, which is 4 bits? maybe because we add 128
			buff8bit[idx] = (BYTE)(pSamples[idx] + 128);
		}
		//write the buffer
		DWORD dwWritten = 0;
		if(WriteFile(hWav, buff8bit, nrSamples*sizeof(BYTE), &dwWritten, NULL) == FALSE)
			return FALSE;
		if(dwWritten != nrSamples*sizeof(BYTE))
			return FALSE;

		return TRUE;
	}
	else
		return FALSE;
}
